package ru.voskhod.smev.client.api.admin.impl;

import org.w3c.dom.Element;
import ru.gov.smev.admserv.admserv._1.*;
import ru.voskhod.smev.client.api.admin.Constants;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.admin.AdminService;
import ru.voskhod.smev.client.api.services.admin.configuration.AdminServiceConfiguration;
import ru.voskhod.smev.client.api.services.identification.IdentityService;
import ru.voskhod.smev.client.api.services.signature.Signer;
import ru.voskhod.smev.client.api.types.exception.SMEVProcessingException;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.exception.processing.SMEVFailureException;
import ru.voskhod.smev.client.api.types.exception.processing.SMEVRoutingException;
import ru.voskhod.smev.client.api.types.exception.processing.SMEVSignatureException;

import java.net.MalformedURLException;

final class AdminServiceImpl implements AdminService {
    public static final String SMEV_VOID_CODE = "-1";
    private static final TransformerServiceLocal transformer = new TransformerServiceLocalImpl();
    private final AdministrativeServicePort port;
    private final Signer signer;
    private final IdentityService identity;

    AdminServiceImpl(AdminServiceConfiguration config) throws MalformedURLException, SMEVRuntimeException {
        this.port = EndPointCreator.createPort(config.getUrl(), config.getTimeout(), Constants.ADMINISTRATIVE_SERVICE);
        this.signer = Factory.getSignerInstance(config.getSignerConfig());
        this.identity = Factory.getIdentityGeneratorInstance(config.getIdentityConfig());
    }

    public boolean switchNode()
            throws SMEVProcessingException, SMEVRuntimeException {
        final SwitchRequest request = EndPointCreator.createRequest(identity, signer, transformer);
        final SwitchResponse switchResponse;
        try {
            switchResponse = port.switchNode(request);
        } catch (SMEVFailure smevFailure) {
            throw new SMEVFailureException(smevFailure.getMessage(), getCode(smevFailure.getFaultInfo()),
                    getDescription(smevFailure.getFaultInfo()), smevFailure);
        } catch (SenderIsNotRegisteredFault ex) {
            throw new SMEVRoutingException(ex.getMessage(), getCode(ex.getFaultInfo()),
                    new SMEVRoutingException.SenderIsNotRegisteredException(ex.getMessage(),
                            getDescription(ex.getFaultInfo()), ex));
        } catch (SenderIsSwitchBlockedFault senderIsSwitchBlockedFault) {
            throw new SMEVRoutingException(senderIsSwitchBlockedFault.getMessage(),
                    getCode(senderIsSwitchBlockedFault.getFaultInfo()),
                    new SMEVRoutingException.AccessDeniedException(senderIsSwitchBlockedFault.getMessage(),
                            getDescription(senderIsSwitchBlockedFault.getFaultInfo()), senderIsSwitchBlockedFault));
        } catch (SignatureVerificationFault_Exception ex) {
            throw new SMEVSignatureException(ex.getMessage(), getCode(ex.getFaultInfo()),
                    new SMEVSignatureException.SignatureVerificationFaultException(ex.getMessage(),
                            getCode(ex.getFaultInfo()), getDescription(ex.getFaultInfo()), ex));
        }
        final XMLDSigSignatureType signature = switchResponse.getSMEVSignature();
        final SwitchResponseInfo responseInfo = switchResponse.getSwitchResponseInfo();
        final Element marshal = transformer.marshal(responseInfo, null);
        signer.validateSMEVSignature(signature.getAny(), marshal);
        return "ACCEPTED".equalsIgnoreCase(switchResponse.getSwitchResponseInfo().getAccepted());
    }

    /**
     * @return
     * @throws SMEVRuntimeException
     * @throws SMEVProcessingException
     */
    @Override
    public String getCurrentNode() throws SMEVRuntimeException, SMEVProcessingException {
        final CurrentNodeRequest request = EndPointCreator.createNodeRequest(identity, signer, transformer);
        final CurrentNodeResponse nodeResponse;
        try {
            nodeResponse = port.getCurrentNode(request);
        } catch (SMEVFailure smevFailure) {
            throw new SMEVFailureException(smevFailure.getMessage(), getCode(smevFailure.getFaultInfo()),
                    getDescription(smevFailure.getFaultInfo()), smevFailure);
        } catch (SenderIsNotRegisteredFault ex) {
            throw new SMEVRoutingException(ex.getMessage(), getCode(ex.getFaultInfo()),
                    new SMEVRoutingException.SenderIsNotRegisteredException(ex.getMessage(),
                            getDescription(ex.getFaultInfo()), ex));
        } catch (SignatureVerificationFault_Exception ex) {
            throw new SMEVSignatureException(ex.getMessage(), getCode(ex.getFaultInfo()),
                    new SMEVSignatureException.SignatureVerificationFaultException(ex.getMessage(),
                            getCode(ex.getFaultInfo()), getDescription(ex.getFaultInfo()), ex));
        }
        final XMLDSigSignatureType signature = nodeResponse.getSMEVSignature();
        final CurrentNodeResponseInfo responseInfo = nodeResponse.getCurrentNodeResponseInfo();
        final Element marshal = transformer.marshal(responseInfo, null);
        signer.validateSMEVSignature(signature.getAny(), marshal);
        return responseInfo.getNode();
    }

    private String getCode(AdministrativeServiceFault aVoid) {
        if (aVoid != null) {
            return aVoid.getCode();
        } else {
            return SMEV_VOID_CODE;
        }
    }

    private String getDescription(AdministrativeServiceFault aVoid) {
        if (aVoid != null) {
            return aVoid.getDescription();
        } else {
            return null;
        }
    }
}
