package ru.voskhod.smev.client.api.poll.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.voskhod.smev.client.api.services.poll.PollInterface;
import ru.voskhod.smev.client.api.services.poll.SyncAdapter;
import ru.voskhod.smev.client.api.services.poll.SyncAdapterConfig;
import ru.voskhod.smev.client.api.services.poll.SyncCallback;

import javax.xml.namespace.QName;
import java.util.*;
import java.util.concurrent.*;

/**
 * The type Poll adapter.
 *
 * @param <T> the type parameter
 */
public class PollAdapter<T> {

    private static final Logger logger = LoggerFactory.getLogger(PollAdapter.class);

    private final PollInterface<T> client;
    private final long checkInterval;
    private final TimeUnit checkIntervalUnit;
    private final long maxTimeout;
    private final TimeUnit maxTimeoutUnit;
    private final ScheduledExecutorService timerExecutor = Executors.newSingleThreadScheduledExecutor();
    private final ExecutorService requestExecutor;
    private final ThreadPoolExecutor pollExecutor;

    private final PollData<T> data = new PollData<>();
    private final Set<PollKey> activeKeys = Collections.newSetFromMap(new HashMap<PollKey, Boolean>());

    /**
     * Instantiates a new Poll adapter.
     *
     * @param client            the client
     * @param checkInterval     the check interval
     * @param checkIntervalUnit the check interval unit
     * @param maxTimeout        the max timeout
     * @param maxTimeoutUnit    the max timeout unit
     * @param requestThreads    the request threads
     * @param pollThreads       the poll threads
     */
    public PollAdapter(PollInterface<T> client,
                       long checkInterval, TimeUnit checkIntervalUnit,
                       long maxTimeout, TimeUnit maxTimeoutUnit,
                       int requestThreads, int pollThreads) {
        if (checkIntervalUnit.toMillis(checkInterval) < 100)
            throw new IllegalArgumentException("Значение checkInterval должно быть не меньше 100 миллисекунд"); // todo: ???
        if (maxTimeoutUnit.toMillis(maxTimeout) > TimeUnit.HOURS.toMillis(1))
            throw new IllegalArgumentException("Максимальный таймаут не должен превышать 1 часа"); // todo: ???
        if (requestThreads > 100)
            throw new IllegalArgumentException("Число requestThreads не должно превышать 100");
        if (pollThreads > 33)
            throw new IllegalArgumentException("Число pollThreads не должно превышать 33");
        this.client = client;
        this.checkInterval = checkInterval;
        this.checkIntervalUnit = checkIntervalUnit;
        this.maxTimeout = maxTimeout;
        this.maxTimeoutUnit = maxTimeoutUnit;
        this.requestExecutor = Executors.newFixedThreadPool(requestThreads);
        this.pollExecutor = new ThreadPoolExecutor(
                pollThreads,
                pollThreads, 0,
                TimeUnit.MILLISECONDS,
                new LinkedBlockingQueue<Runnable>());
        startPoll();
    }

    private void startPoll() {
        timerExecutor.scheduleWithFixedDelay(new Runnable() {
            @Override
            public void run() {
                List<PollKey<T>> keys = data.listKeys();
                // Предупреждение переполнения очереди:
                if (pollExecutor.getActiveCount() < pollExecutor.getCorePoolSize()) {
                    // Есть свободные потоки, валим все задачи опроса в executor, но сначала - для тех типов,
                    // которые пока не обрабатываются (есть шанс, что обрабатывающиеся типы скоро завершат работу
                    // и вернут значения для этих типов в запросы, так что еще раз добавлять их излишне):
                    for (Iterator<PollKey<T>> i = keys.iterator(); i.hasNext(); ) {
                        PollKey<T> key = i.next();
                        if (activeKeys.add(key)) {
                            pollKeyAsync(key);
                            i.remove();
                        }
                    }
                    // Потом добавляем оставшиеся типы (которые уже обрабатываются).
                    // Если обрабатывающиеся типы по каким-то причинам зависли надолго, то можно надеяться, что
                    // добавление этих же типов еще раз проскочит быстрее при опросе - свободные потоки для выполнения
                    // пока есть:
                    for (PollKey<T> key : keys) {
                        activeKeys.add(key);
                        pollKeyAsync(key);
                    }
                } else {
                    // Если все потоки заняты, то добавляем в очередь только те типы, которые еще не обрабатываются.
                    // Таким образом, если задачи опроса зависли, и новые задачи помещаются в очередь,
                    // то максимальная длина очереди - кол-во типов запросов (больше все равно не имеет смысла):
                    for (PollKey<T> key : keys) {
                        if (activeKeys.add(key)) {
                            pollKeyAsync(key);
                        }
                    }
                }
            }
        }, checkInterval, checkInterval, checkIntervalUnit);
    }

    /**
     * Gets check interval.
     *
     * @return the check interval
     */
    public long getCheckInterval() {
        return checkInterval;
    }

    /**
     * Gets check interval unit.
     *
     * @return the check interval unit
     */
    public TimeUnit getCheckIntervalUnit() {
        return checkIntervalUnit;
    }

    /**
     * Gets timer executor.
     *
     * @return the timer executor
     */
    public ScheduledExecutorService getTimerExecutor() {
        return timerExecutor;
    }

    /**
     * Gets request executor.
     *
     * @return the request executor
     */
    public ExecutorService getRequestExecutor() {
        return requestExecutor;
    }

    /**
     * Gets poll executor.
     *
     * @return the poll executor
     */
    public ExecutorService getPollExecutor() {
        return pollExecutor; // todo: protect from changing of number of threads
    }

    /**
     * Sync request t.
     *
     * @param messageId   the message id
     * @param type        the type
     * @param nodeId      the node id
     * @param timeout     если null, то использовать maxTimeout, указанный при создании адаптера
     * @param timeoutUnit используется, только если timeout - не null (если timeout == null, то может быть любое                    значение, включая null)
     * @param sendRequest the send request
     * @return the t
     * @throws InterruptedException the interrupted exception
     * @throws ExecutionException   the execution exception
     * @throws TimeoutException     the timeout exception
     */
    public T syncRequest(final String messageId, QName type, String nodeId, Long timeout, TimeUnit timeoutUnit,
                         final Callable<? super Void> sendRequest, final SyncCallback<T> callback)
            throws InterruptedException, ExecutionException, TimeoutException {
        if (timeout != null) {
            if (timeoutUnit.toMillis(timeout) > maxTimeoutUnit.toMillis(maxTimeout))
                throw new IllegalArgumentException("Таймаут не должен превышать " + maxTimeout + " " + maxTimeoutUnit);
        }
        if (timeout == null) {
            timeout = maxTimeout;
            timeoutUnit = maxTimeoutUnit;
        }
        PollKey<T> key = new PollKey<>(type, nodeId, callback);
        final ResponseTask<T> task = new ResponseTask<>(messageId, key);
        data.addTask(task);
        logger.debug("Запрос поставлен в очередь опроса СМЭВ: messageId={}, type={}, nodeId={}, timeout={} {}", messageId, type, nodeId, timeout, timeoutUnit);
        Future<?> future = requestExecutor.submit(new Runnable() {
            @Override
            public void run() {
                try {
                    sendRequest.call();
                    callback.event(SyncAdapter.Event.SUCCESS);
                } catch (Throwable ex) {
                    logger.error("Ошибка при отправке запроса СМЭВ: messageId={}", messageId, ex);
                    callback.event(SyncAdapter.Event.FAILURE);
                    cancel(task, ex);
                }
            }
        });
        try {
            return task.get(timeout, timeoutUnit);
        } catch (TimeoutException ex) {
            TimeoutException tex = new TimeoutException("Превышение пороговой продолжительности обработки вызова");
            tex.setStackTrace(ex.getStackTrace());
            throw tex;
        } finally {
            cancel(task, null);
            future.cancel(false);
        }
    }

    private void cancel(ResponseTask<T> task, Throwable error) {
        data.cancelTask(task);
        if (error != null) {
            task.setError(error);
        }
    }

    private void pollKeyAsync(final PollKey<T> key) {
        pollExecutor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    pollKey(key);
                } finally {
                    activeKeys.remove(key);
                }
            }
        });
    }

    private void pollKey(PollKey<T> key) {
        logger.debug("Опрос ответов СМЭВ для type={}, nodeId={}", key.type, key.nodeId);
        try {
            while (true) {
                final SyncCallback<T> callback = key.callback;
                T response = callback.getResponse();
                if (!processResponse(response, callback))
                    break;
                response = client.getResponse(key.type, key.nodeId);
                if (!processResponse(response, callback))
                    break;
            }
        } catch (InterruptedException ex) {
            Thread.currentThread().interrupt();
        } catch (Exception ex) {
            logger.error("Ошибка при получении ответа СМЭВ", ex);
        }
    }

    private boolean processResponse(T response, SyncCallback<T> callback) {
        if (response == null)
            return false;
        callback.received(response);
        String messageId = client.getMessageId(response);
        if (messageId == null)
            return false;
        logger.debug("При опросе СМЭВ получен ответ на сообщение {}", messageId);
        ResponseTask<T> task = data.removeTask(messageId);
        if (task != null) {
            task.setResult(response);
        }
        return true;
    }

    /**
     * Shutdown now.
     */
    public void shutdownNow() {
        timerExecutor.shutdownNow();
        requestExecutor.shutdownNow();
        pollExecutor.shutdownNow();
    }

    /**
     * Gets waiting count.
     *
     * @return the waiting count
     */
    public int getWaitingCount() {
        return data.getWaitingCount();
    }

    /**
     * Create adapter poll adapter.
     *
     * @param <T>    the type parameter
     * @param client the client
     * @return the poll adapter
     */

    public static <T> PollAdapter<T> createAdapter(PollInterface<T> client, SyncAdapterConfig config) {
        return new PollAdapter<>(
                client,
                config.getCheckInterval(), config.getCheckIntervalUnit(),
                config.getMaxTimeout(), config.getMaxTimeoutUnit(),
                config.getRequestThreads(), config.getPollThreads()
        );
    }
}
