package ru.voskhod.smev.client.api.poll.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * The type Poll data.
 *
 * @param <T> the type parameter
 */
final class PollData<T> {

    private static final Logger logger = LoggerFactory.getLogger(PollData.class);
    private final Map<PollKey<T>, Counter> keys = new HashMap<>();
    private final Map<String, ResponseTask<T>> waiting = new HashMap<>();
    private final Object dataLock = new Object();

    /**
     * Add task.
     *
     * @param task the task
     */
    void addTask(ResponseTask<T> task) {
        String messageId = task.messageId;
        PollKey<T> key = task.key;
        synchronized (dataLock) {
            ResponseTask<T> oldTask = waiting.put(messageId, task);
            if (oldTask != null) {
                removeKey(oldTask);
            }
            Counter counter = keys.get(key);
            if (counter == null) {
                counter = new Counter();
                keys.put(key, counter);
            }
            counter.add(task);
        }
        logger.debug("Добавлен запрос в очередь: messageId={}, type={}, nodeId={}", messageId, key.type, key.nodeId);
    }

    /**
     * Remove task response task.
     *
     * @param messageId the message id
     * @return the response task
     */
    ResponseTask<T> removeTask(String messageId) {
        ResponseTask<T> task;
        synchronized (dataLock) {
            task = waiting.remove(messageId);
            if (task != null) {
                removeKey(task);
            }
        }
        logger.debug("Удален запрос из очереди: messageId={}", messageId);
        return task;
    }

    /**
     * Cancel task boolean.
     *
     * @param task the task
     * @return the boolean
     */
    boolean cancelTask(ResponseTask<T> task) {
        synchronized (dataLock) {
            ResponseTask<T> waitingTask = waiting.get(task.messageId);
            if (waitingTask != task)
                return false;
            waiting.remove(task.messageId);
            removeKey(task);
        }
        logger.debug("Прерван запрос из очереди: messageId={}", task.messageId);
        return true;
    }

    private void removeKey(ResponseTask<T> task) {
        Counter counter = keys.get(task.key);
        if (counter != null) {
            if (counter.remove(task)) {
                keys.remove(task.key);
            }
        }
    }

    /**
     * List keys list.
     *
     * @return список типов запросов, отсортированный по возрастанию среднего времени создания запроса этого типа
     */
    List<PollKey<T>> listKeys() {
        List<KeyWithTimestamp<T>> tt;
        synchronized (dataLock) {
            if (keys.isEmpty()) {
                return Collections.emptyList();
            } else {
                tt = new ArrayList<>(keys.size());
                for (Map.Entry<PollKey<T>, Counter> entry : keys.entrySet()) {
                    PollKey<T> key = entry.getKey();
                    Counter counter = entry.getValue();
                    tt.add(new KeyWithTimestamp<>(key, counter.getTimestamp()));
                }
            }
        }
        Collections.sort(tt, new Comparator<KeyWithTimestamp>() {
            public int compare(KeyWithTimestamp o1, KeyWithTimestamp o2) {
                return Long.compare(o1.timestamp, o2.timestamp);
            }
        });
        List<PollKey<T>> result = new ArrayList<>(tt.size());
        for (KeyWithTimestamp<T> t : tt) {
            result.add(t.key);
        }
        return result;
    }

    /**
     * Gets waiting count.
     *
     * @return the waiting count
     */
    int getWaitingCount() {
        synchronized (dataLock) {
            int sum = 0;
            for (Counter counter : keys.values()) {
                sum += counter.count;
            }
            if (sum != waiting.size())
                throw new IllegalStateException("Число типов: " + sum + ", число сообщений: " + waiting.size());
            return waiting.size();
        }
    }

    private static final class Counter {

        private long sumTimestamp = 0;
        private int count = 0;

        /**
         * Add.
         *
         * @param task the task
         */
        void add(ResponseTask<?> task) {
            sumTimestamp += task.timestamp;
            count++;
        }

        /**
         * Remove boolean.
         *
         * @param task the task
         * @return the boolean
         */
        boolean remove(ResponseTask<?> task) {
            sumTimestamp -= task.timestamp;
            count--;
            return count <= 0;
        }

        /**
         * Gets timestamp.
         *
         * @return the timestamp
         */
        long getTimestamp() {
            return sumTimestamp / count;
        }
    }

    private static final class KeyWithTimestamp<T> {

        /**
         * The Key.
         */
        final PollKey<T> key;
        /**
         * The Timestamp.
         */
        final long timestamp;

        private KeyWithTimestamp(PollKey<T> key, long timestamp) {
            this.key = key;
            this.timestamp = timestamp;
        }
    }
}
