package ru.voskhod.smev.client.api.template.impl;

import it.sauronsoftware.ftp4j.FTPAbortedException;
import it.sauronsoftware.ftp4j.FTPClient;
import it.sauronsoftware.ftp4j.FTPDataTransferException;
import it.sauronsoftware.ftp4j.FTPException;
import it.sauronsoftware.ftp4j.FTPIllegalReplyException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.voskhod.smev.client.api.services.transport.LargeAttachmentTransport;
import ru.voskhod.smev.client.api.services.transport.configuration.LargeAttachmentTransportConfiguration;
import ru.voskhod.smev.client.api.types.exception.SMEVException;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.message.attachment.LargeAttachment;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;


/**
 * The type Large attachment transport.
 */
public class LargeAttachmentTransportImpl implements LargeAttachmentTransport {
    private static final Logger logger = LoggerFactory.getLogger(LargeAttachmentTransportImpl.class);
    private int maxAttempts = 3;
    private String ftpLogin = "anonymous";
    private String ftpPass = "smev";

    private String ftpAddress;
    private FTPClient ftpClient;

    /**
     * Instantiates a new Large attachment transport.
     *
     * @param config the config
     */
    public LargeAttachmentTransportImpl(LargeAttachmentTransportConfiguration config) {
        this.maxAttempts = config.getMaxAttempts();
        this.ftpAddress = config.getAddress();
        this.ftpClient = new FTPClient();
        this.ftpLogin = config.getLogin();
        this.ftpPass = config.getPass();
    }

    @Override
    public synchronized boolean download(LargeAttachment attachment, OutputStream output) throws SMEVException {
        try {
            logger.debug("Соединяемся");
            connect();
            boolean ok = false;
            try {
                LargeAttachment.TransportDetails transportDetails = attachment.getTransportDetails();
                logger.debug("Логин");
                ftpClient.login(transportDetails.getUserName(), transportDetails.getPassword());
                logger.debug("Загрузка");
                ftpClient.download(attachment.getContentRef().getPath(), output, 0, null);
                ok = true;
                return true;
            } finally {
                logger.debug("Отсоединение");
                ftpClient.disconnect(ok);
            }
        } catch (FTPAbortedException ex) {
            return false;
        } catch (FTPIllegalReplyException | FTPDataTransferException | FTPException | IOException ex) {
            throw new SMEVRuntimeException(ex.getMessage(), ex);
        }
    }

    @Override
    public synchronized void upload(LargeAttachment attachment) throws SMEVRuntimeException {
        for (int attempt = 0; attempt < maxAttempts; attempt++) {
            try {
                //  FTPClient ftp = new FTPClient();
                logger.debug("Соединяемся");
                connect();
                boolean ok = false;
                try {
                    logger.debug("Логин");
                    ftpClient.login(ftpLogin, ftpPass);
                    logger.debug("Создание директории");
                    ftpClient.createDirectory(attachment.getUuid().toString());
                    ftpClient.changeDirectory(attachment.getUuid().toString());
                    logger.debug("Аплоад");
                    ftpClient.upload(new File(attachment.getContentRef()));
                    ok = true;
                } finally {
                    logger.debug("Отсоединение");
                    ftpClient.disconnect(ok);
                }
                if (ok) {
                    return;
                }
            } catch (FTPException | FTPIllegalReplyException | IOException | FTPDataTransferException | FTPAbortedException ex) {
                if (attempt >= maxAttempts - 1) {
                    throw new SMEVRuntimeException(ex.getMessage(), ex);
                }
            }
        }
        throw new IllegalStateException();
    }

    @Override
    public void cancel() {
        if (ftpClient.isConnected()) {
            try {
                ftpClient.abortCurrentDataTransfer(true);
            } catch (Exception ex) {
                try {
                    ftpClient.abortCurrentDataTransfer(false);
                } catch (Exception ignored) {
                }
            }
        } else {
            ftpClient.abortCurrentConnectionAttempt();
        }
    }

    private void connect() throws FTPException, IOException, FTPIllegalReplyException {
        String[] addressSplit = ftpAddress.split(":");
        switch (addressSplit.length) {
            case 1:
                ftpClient.connect(addressSplit[0]);
                break;
            case 2:
                ftpClient.connect(addressSplit[0], Integer.parseInt(addressSplit[1]));
                break;
        }
    }
}
