package ru.voskhod.smev.client.api.template.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.identification.IdentityService;
import ru.voskhod.smev.client.api.services.signature.Signer;
import ru.voskhod.smev.client.api.services.template.LargeAttachmentTemplate;
import ru.voskhod.smev.client.api.services.template.MTOMAttachmentTemplate;
import ru.voskhod.smev.client.api.services.template.WSTemplate;
import ru.voskhod.smev.client.api.services.template.configuration.GeoTemplateConfiguration;
import ru.voskhod.smev.client.api.services.template.configuration.TemplateConfiguration;
import ru.voskhod.smev.client.api.services.transport.LargeAttachmentTransport;
import ru.voskhod.smev.client.api.signature.impl.SignerFactory;
import ru.voskhod.smev.client.api.types.exception.SMEVException;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.message.SMEVMessage;
import ru.voskhod.smev.client.api.types.message.attachment.SMEVAttachment;
import ru.voskhod.smev.client.api.types.message.business.data.BusinessContent;
import ru.voskhod.smev.client.api.types.message.business.data.request.RequestContent;
import ru.voskhod.smev.client.api.types.message.business.data.response.RejectResponseContent;
import ru.voskhod.smev.client.api.types.message.business.data.response.ResponseContent;
import ru.voskhod.smev.client.api.types.message.business.data.response.StatusResponseContent;
import ru.voskhod.smev.client.api.types.message.system.SMEVMetadata;
import ru.voskhod.smev.client.api.types.message.system.processing.ProcessingInformation;
import ru.voskhod.smev.client.api.types.message.system.processing.QueryInformation;
import ru.voskhod.smev.client.api.types.message.system.processing.RequestInformation;
import ru.voskhod.smev.client.api.types.message.system.processing.ResponseInformation;
import ru.voskhod.smev.client.api.util.Utils;

import javax.xml.namespace.QName;
import java.io.File;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * The type Template.
 */
public class TemplateImpl implements ru.voskhod.smev.client.api.services.template.Template {
    private static final Logger logger = LoggerFactory.getLogger(TemplateImpl.class);

    private Signer signer;
    private IdentityService identificationService;
    private MTOMAttachmentTemplate mtomAttachmentTemplate;
    private LargeAttachmentTemplate largeAttachmentTemplate;
    private WSTemplate wsTemplate;
    /**
     * The Large attachment transport.
     */
    LargeAttachmentTransport largeAttachmentTransport;

    /**
     * Instantiates a new Template.
     *
     * @param templateConfiguration the template configuration
     * @param cert                  the cert
     * @param privateKey            the private key
     * @param pass                  the pass
     * @throws SMEVRuntimeException the smev runtime exception
     */
    public TemplateImpl(TemplateConfiguration templateConfiguration, String cert, String privateKey, String pass) throws SMEVRuntimeException {
        this.signer = SignerFactory.getSigner(templateConfiguration.getSignerConfig(), cert, privateKey, pass);
        this.identificationService = Factory.getIdentityGeneratorInstance(templateConfiguration.getIdentityConfig());

        if (templateConfiguration.getLargeAttachmentTransportConfig() != null) {
            logger.debug("Инициализируем шаблон загрузки больших файлов");
            largeAttachmentTransport = Factory.getLargeAttachmentTransportInstance(templateConfiguration.getLargeAttachmentTransportConfig());
        }

        this.largeAttachmentTemplate = Factory.getLargeAttachmentTemplateInstance(signer, identificationService, largeAttachmentTransport);

        mtomAttachmentTemplate = Factory.getMTOMAttachmentTemplateInstance(signer, identificationService);

        wsTemplate = Factory.getWsTemplateInstance(signer,
                Factory.getValidatorInstance(templateConfiguration.getValidatorConfig()),
                Factory.getMessageMapperInstance(templateConfiguration.getMessageGenerationConfig()),
                Factory.getMessageTransportInstance(templateConfiguration.getMessageTransportConfiguration()),
                Factory.getExceptionMapperInstance(templateConfiguration.getMessageGenerationConfig()),
                largeAttachmentTransport,
                templateConfiguration
        );
    }

    public TemplateImpl(GeoTemplateConfiguration templateConfiguration, String cert, String privateKey, String pass) throws SMEVRuntimeException {
        this.signer = SignerFactory.getSigner(templateConfiguration.getSignerConfig(), cert, privateKey, pass);
        this.identificationService = Factory.getIdentityGeneratorInstance(templateConfiguration.getIdentityConfig());

        if (templateConfiguration.getLargeAttachmentTransportConfig() != null) {
            logger.debug("Инициализируем шаблон загрузки больших файлов");
            largeAttachmentTransport = Factory.getLargeAttachmentTransportInstance(templateConfiguration.getLargeAttachmentTransportConfig());
        }

        this.largeAttachmentTemplate = Factory.getLargeAttachmentTemplateInstance(signer, identificationService, largeAttachmentTransport);

        mtomAttachmentTemplate = Factory.getMTOMAttachmentTemplateInstance(signer, identificationService);

        wsTemplate = Factory.getWsTemplateInstance(signer,
                Factory.getValidatorInstance(templateConfiguration.getValidatorConfig()),
                Factory.getMessageMapperInstance(templateConfiguration.getMessageGenerationConfig()),
                Factory.getMessageTransportInstance(templateConfiguration.getGeoMessageTransportConfiguration()),
                Factory.getExceptionMapperInstance(templateConfiguration.getMessageGenerationConfig()),
                largeAttachmentTransport,
                templateConfiguration
        );
    }

    @Override
    public SMEVMetadata sendRequestMessage(String nodeId,
                                           String messageId,
                                           String transactionCode,
                                           String originalMessageId,
                                           String referenceMessageId, List<Element> bpm,
                                           Date eol,
                                           Element content, Element personalSignature,
                                           List<File> mtomAttachments,
                                           List<File> largeAttachments, Map<String, String> extendedProps,
                                           boolean testMessage) throws SMEVException {
        if (messageId == null) {
            logger.info("Создаем messageId");
            messageId = identificationService.generateUUID();
            referenceMessageId = messageId;
        } else {
            if (referenceMessageId == null) {
                logger.info("ReferenceId=MessageId");
                referenceMessageId = messageId;
            }
        }


        SMEVMetadata.MessageIdentity messageIdentity = new SMEVMetadata.MessageIdentity(messageId, referenceMessageId, transactionCode);
        ProcessingInformation processingInformation = new RequestInformation(eol != null ? Utils.getXmlTimestamp(eol) : null, nodeId, testMessage);

        SMEVMetadata smevMetadata = new SMEVMetadata(messageIdentity, processingInformation);//todo:extended
        List<SMEVAttachment> smevAttachments = new ArrayList<>();
        logger.debug("Добавляем MTOM приложения");
        smevAttachments.addAll(mtomAttachmentTemplate.createAttachments(mtomAttachments));
        logger.debug("Добавляем ftp приложения");
        smevAttachments.addAll(largeAttachmentTemplate.createAttachments(largeAttachments));

        BusinessContent businessData = new RequestContent(content, personalSignature, smevAttachments, bpm);
        logger.debug("Создаем сообщение");
        SMEVMessage smevMessage = new SMEVMessage(smevMetadata, businessData);
        logger.debug("Посылаем сообщение");
        return wsTemplate.send(smevMessage);
    }

    @Override
    public SMEVMetadata sendResponseMessage(
            String messageId,
            String replyToAddress,
            String originalMessageId,
            List<Element> bpm,
            Element content, Element personalSignature,
            List<File> mtomAttachments,
            List<File> largeAttachments, Map<String, String> extendedProps) throws SMEVException {
        if (messageId == null) {
            logger.info("Создаем messageId");
            messageId = identificationService.generateUUID();
        }


        SMEVMetadata.MessageIdentity messageIdentity = new SMEVMetadata.MessageIdentity(messageId, null, null);
        ProcessingInformation processingInformation = new ResponseInformation(originalMessageId, replyToAddress);

        SMEVMetadata smevMetadata = new SMEVMetadata(messageIdentity, processingInformation);
        List<SMEVAttachment> smevAttachments = new ArrayList<>();
        logger.debug("Добавляем MTOM приложения");
        smevAttachments.addAll(mtomAttachmentTemplate.createAttachments(mtomAttachments));
        logger.debug("Добавляем ftp приложения");
        smevAttachments.addAll(largeAttachmentTemplate.createAttachments(largeAttachments));


        BusinessContent businessData = new ResponseContent(content, personalSignature, smevAttachments, bpm);
        logger.debug("Создаем сообщение");
        SMEVMessage smevMessage = new SMEVMessage(smevMetadata, businessData);
        logger.debug("Посылаем сообщение");
        return wsTemplate.send(smevMessage);
    }

    @Override
    public SMEVMetadata sendRejectResponseMessage(
            String messageId,
            String replyToAddress,
            String originalMessageId,
            String code, String description, List<Element> bpm
    ) throws SMEVException {
        if (messageId == null) {
            logger.info("Создаем messageId");
            messageId = identificationService.generateUUID();
        }


        SMEVMetadata.MessageIdentity messageIdentity = new SMEVMetadata.MessageIdentity(messageId, null, null);

        ProcessingInformation processingInformation = new ResponseInformation(originalMessageId, replyToAddress);

        SMEVMetadata smevMetadata = new SMEVMetadata(messageIdentity, processingInformation);

        final RejectResponseContent.RejectCode rejectCode = RejectResponseContent.RejectCode.fromValue(code);
        RejectResponseContent businessData = new RejectResponseContent(bpm);
        businessData.add(rejectCode, description);
        logger.debug("Создаем сообщение");
        SMEVMessage smevMessage = new SMEVMessage(smevMetadata, businessData);
        logger.debug("Посылаем сообщение");
        return wsTemplate.send(smevMessage);
    }

    @Override
    public SMEVMetadata sendStatusResponseMessage(
            String messageId,
            String replyToAddress,
            String originalMessageId,
            int code, String description, Map<String, String> map, List<Element> bpm
    ) throws SMEVException {
        if (messageId == null) {
            logger.info("Создаем messageId");
            messageId = identificationService.generateUUID();
        }


        SMEVMetadata.MessageIdentity messageIdentity = new SMEVMetadata.MessageIdentity(messageId, null, null);

        ProcessingInformation processingInformation = new ResponseInformation(originalMessageId, replyToAddress);

        SMEVMetadata smevMetadata = new SMEVMetadata(messageIdentity, processingInformation);

        BusinessContent meta = new StatusResponseContent(code, description, map, bpm);
        logger.debug("Создаем сообщение");
        SMEVMessage smevMessage = new SMEVMessage(smevMetadata, meta);
        logger.debug("Посылаем сообщение");
        return wsTemplate.send(smevMessage);
    }


    @Override
    public SMEVMessage getRequestMessage(QName requestTypeSelector, String nodeId) throws SMEVException {
        QueryInformation.Type type = QueryInformation.Type.REQUEST;
        logger.debug("Создаем запрос");
        QueryInformation queryInformation = new QueryInformation(nodeId, requestTypeSelector == null ? null : requestTypeSelector.getNamespaceURI(), requestTypeSelector == null ? null : requestTypeSelector.getLocalPart(), type);
        logger.debug("Посылаем сообщение");
        return wsTemplate.get(queryInformation);
    }

    @Override
    public SMEVMessage getStatusMessage(QName requestTypeSelector, String nodeId) throws SMEVException {
        QueryInformation.Type type = QueryInformation.Type.STATUS;
        logger.debug("Создаем запрос");
        QueryInformation queryInformation = new QueryInformation(nodeId, requestTypeSelector == null ? null : requestTypeSelector.getNamespaceURI(), requestTypeSelector == null ? null : requestTypeSelector.getLocalPart(), type);
        logger.debug("Получаем сообщение");
        return wsTemplate.get(queryInformation);
    }

    @Override
    public SMEVMessage getResponseMessage(QName requestTypeSelector, String nodeId) throws SMEVException {
        QueryInformation.Type type = QueryInformation.Type.RESPONSE;
        logger.debug("Создаем запрос");
        QueryInformation queryInformation = new QueryInformation(nodeId, requestTypeSelector == null ? null : requestTypeSelector.getNamespaceURI(), requestTypeSelector == null ? null : requestTypeSelector.getLocalPart(), type);
        logger.debug("Получаем сообщение");
        return wsTemplate.get(queryInformation);
    }


    @Override
    public void ack(String messageId)
            throws SMEVException {
        logger.debug("Создаем запрос");
        SMEVMetadata smevMetadata = new SMEVMetadata(new SMEVMetadata.MessageIdentity(messageId, null, null), null);
        logger.debug("Посылаем сообщение");
        wsTemplate.ack(smevMetadata, true);
    }

    @Override
    public Map<String, Long> getIncomingQueueStatistics(String nodeId)
            throws SMEVException {
        logger.debug("Получаем сообщение");
        return wsTemplate.getIncomingQueueStatistics(nodeId);
    }
}
