package ru.voskhod.smev.client.api.template.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.messaging.ExceptionMapper;
import ru.voskhod.smev.client.api.services.messaging.MessageMapper;
import ru.voskhod.smev.client.api.services.signature.PostValidationCallback;
import ru.voskhod.smev.client.api.services.signature.Signer;
import ru.voskhod.smev.client.api.services.template.WSTemplate;
import ru.voskhod.smev.client.api.services.template.configuration.GeoTemplateConfiguration;
import ru.voskhod.smev.client.api.services.template.configuration.WSTemplateConfiguration;
import ru.voskhod.smev.client.api.services.transport.GeoMessageTransport;
import ru.voskhod.smev.client.api.services.transport.LargeAttachmentTransport;
import ru.voskhod.smev.client.api.services.transport.MessageTransport;
import ru.voskhod.smev.client.api.services.validation.Validator;
import ru.voskhod.smev.client.api.template.marshaller.Marshaller;
import ru.voskhod.smev.client.api.types.exception.SMEVException;
import ru.voskhod.smev.client.api.types.exception.SMEVProcessingException;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.exception.processing.SMEVSignatureException;
import ru.voskhod.smev.client.api.types.message.SMEVMessage;
import ru.voskhod.smev.client.api.types.message.attachment.LargeAttachment;
import ru.voskhod.smev.client.api.types.message.attachment.MTOMAttachment;
import ru.voskhod.smev.client.api.types.message.attachment.SMEVAttachment;
import ru.voskhod.smev.client.api.types.message.business.data.request.AbstractRequestContent;
import ru.voskhod.smev.client.api.types.message.business.data.response.AbstractResponseContent;
import ru.voskhod.smev.client.api.types.message.business.enums.BusinessStatus;
import ru.voskhod.smev.client.api.types.message.system.SMEVContext;
import ru.voskhod.smev.client.api.types.message.system.SMEVMetadata;
import ru.voskhod.smev.client.api.types.message.system.processing.QueryInformation;
import ru.voskhod.smev.client.api.types.message.system.processing.RequestInformation;
import ru.voskhod.smev.client.api.types.message.system.processing.ResponseInformation;
import ru.voskhod.smev.client.api.types.message.transport.GeoMessage;
import ru.voskhod.smev.server.autogenerated.types.*;
import ru.voskhod.smev.server.autogenerated.types.basic.SmevFault;
import ru.voskhod.smev.server.autogenerated.types.basic.Void;
import ru.voskhod.smev.server.autogenerated.types.basic.XMLDSigSignatureType;

import javax.xml.transform.TransformerException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;


/**
 * The type Ws template.
 */
//FIXME: Здесь добавлять функционал резервного СМЭВ
public class WSTemplateImpl implements WSTemplate {

    private static final Logger logger = LoggerFactory.getLogger(WSTemplateImpl.class);
    private Signer signer;
    private Validator validator;
    private MessageMapper<Void, AckRequest, GetStatusResponse, GetStatusRequest, Object,
            Object, GetResponseResponse, GetResponseRequest, SendResponseResponse, SendResponseRequest,
            GetRequestResponse, GetRequestRequest, SendRequestResponse, SendRequestRequest> messageMapper;
    private MessageTransport<SendRequestRequest, SendRequestResponse, SendResponseRequest, SendResponseResponse,
            GetRequestRequest, GetRequestResponse, GetResponseRequest, GetResponseResponse, GetStatusRequest, GetStatusResponse, AckRequest, Void,
            Object, Object> transport;
    private GeoMessageTransport<SendRequestRequest, SendRequestResponse, SendResponseRequest, SendResponseResponse,
            GetRequestRequest, GetRequestResponse, GetResponseRequest, GetResponseResponse, GetStatusRequest, GetStatusResponse, AckRequest, Void,
            Object, Object> geoTransport;
    private ExceptionMapper<SmevFault> exceptionMapper;
    private LargeAttachmentTransport largeAttachmentTransport;

    //todo:   config
    private boolean loggingOutput;
    private boolean validateOutput;
    private boolean validateInput;
    private boolean loggingInput;
    private boolean validateSMEVSignature;

    private String localStorage;

    private final PostValidationCallback postValidationCallback;

    /**
     * Instantiates a new Ws template.
     *
     * @param configuration the configuration
     * @throws SMEVRuntimeException the smev runtime exception
     */
    public WSTemplateImpl(WSTemplateConfiguration configuration) throws SMEVRuntimeException {

        signer = Factory.getSignerInstance(configuration.getSignerConfig());    //todo:

        if (configuration.getValidatorConfig() != null) {
            logger.debug("Initializing validator");
            validator = Factory.getValidatorInstance(configuration.getValidatorConfig());
        }

        messageMapper = Factory.getMessageMapperInstance(configuration.getMessageGenerationConfig());

        transport = Factory.getMessageTransportInstance(configuration.getMessageTransportConfiguration());

        exceptionMapper = Factory.getExceptionMapperInstance(configuration.getMessageGenerationConfig());

        if (configuration.getLargeAttachmentTransportConfig() != null) {
            logger.debug("Initializing large attachment transport");
            largeAttachmentTransport = Factory.getLargeAttachmentTransportInstance(configuration.getLargeAttachmentTransportConfig());
        }
        //todo:
        loggingInput = configuration.getLoggingInput();
        loggingOutput = configuration.getLoggingOutput();
        validateInput = configuration.getValidateInput();
        validateOutput = configuration.getValidateOutput();
        validateSMEVSignature = configuration.getValidateSMEVSignature();
        localStorage = configuration.getLocalStorage();
        postValidationCallback = null;
    }

    /**
     * Instantiates a new Ws template.
     *
     * @param signer     the signer
     * @param validator  the validator
     * @param mapper     the mapper
     * @param transport  the transport
     * @param manager    the manager
     * @param atransport the atransport
     * @throws SMEVRuntimeException the smev runtime exception
     */
    public WSTemplateImpl(Signer signer,
                          Validator validator,
                          MessageMapper mapper,
                          MessageTransport transport,
                          ExceptionMapper manager,
                          LargeAttachmentTransport atransport,
                          WSTemplateConfiguration configuration) throws SMEVRuntimeException {

        this.signer = signer;

        this.validator = validator;

        this.messageMapper = mapper;

        this.transport = transport;

        this.exceptionMapper = manager;

        this.largeAttachmentTransport = atransport;
        loggingInput = configuration.getLoggingInput();
        loggingOutput = configuration.getLoggingOutput();
        validateInput = configuration.getValidateInput();
        validateOutput = configuration.getValidateOutput();
        validateSMEVSignature = configuration.getValidateSMEVSignature();
        localStorage = configuration.getLocalStorage();
        postValidationCallback = null;
    }

    public WSTemplateImpl(Signer signer,
                          Validator validator,
                          MessageMapper mapper,
                          MessageTransport transport,
                          ExceptionMapper manager,
                          LargeAttachmentTransport atransport,
                          WSTemplateConfiguration configuration,
                          PostValidationCallback postValidationCallback) throws SMEVRuntimeException {

        this.signer = signer;

        this.validator = validator;

        this.messageMapper = mapper;

        this.transport = transport;

        this.exceptionMapper = manager;

        this.largeAttachmentTransport = atransport;
        loggingInput = configuration.getLoggingInput();
        loggingOutput = configuration.getLoggingOutput();
        validateInput = configuration.getValidateInput();
        validateOutput = configuration.getValidateOutput();
        validateSMEVSignature = configuration.getValidateSMEVSignature();
        localStorage = configuration.getLocalStorage();
        this.postValidationCallback = postValidationCallback;
    }

    /**
     * Instantiates a new Ws template.
     *
     * @param signer       the signer
     * @param validator    the validator
     * @param mapper       the mapper
     * @param geoTransport the transport
     * @param manager      the manager
     * @param atransport   the atransport
     * @throws SMEVRuntimeException the smev runtime exception
     */
    public WSTemplateImpl(Signer signer,
                          Validator validator,
                          MessageMapper mapper,
                          GeoMessageTransport geoTransport,
                          ExceptionMapper manager,
                          LargeAttachmentTransport atransport,
                          GeoTemplateConfiguration configuration) throws SMEVRuntimeException {

        this.signer = signer;

        this.validator = validator;

        this.messageMapper = mapper;

        this.geoTransport = geoTransport;

        this.exceptionMapper = manager;

        this.largeAttachmentTransport = atransport;
        loggingInput = configuration.getLoggingInput();
        loggingOutput = configuration.getLoggingOutput();
        validateInput = configuration.getValidateInput();
        validateOutput = configuration.getValidateOutput();
        validateSMEVSignature = configuration.getValidateSMEVSignature();
        localStorage = configuration.getLocalStorage();
        postValidationCallback = null;
    }

    public WSTemplateImpl(Signer signer,
                          Validator validator,
                          MessageMapper mapper,
                          GeoMessageTransport geoTransport,
                          ExceptionMapper manager,
                          LargeAttachmentTransport atransport,
                          GeoTemplateConfiguration configuration,
                          PostValidationCallback postValidationCallback) throws SMEVRuntimeException {

        this.signer = signer;

        this.validator = validator;

        this.messageMapper = mapper;

        this.geoTransport = geoTransport;

        this.exceptionMapper = manager;

        this.largeAttachmentTransport = atransport;
        loggingInput = configuration.getLoggingInput();
        loggingOutput = configuration.getLoggingOutput();
        validateInput = configuration.getValidateInput();
        validateOutput = configuration.getValidateOutput();
        validateSMEVSignature = configuration.getValidateSMEVSignature();
        localStorage = configuration.getLocalStorage();
        this.postValidationCallback = postValidationCallback;
    }

    @Override
    public SMEVMessage get(QueryInformation query) throws SMEVException {
        switch (query.getType()) {
            case REQUEST:
                return getRequest(query);
            case RESPONSE:
                return getResponse(query);
            case STATUS:
                return getStatus();
            default:
                throw new SMEVRuntimeException("Unknown query type!");
        }
    }

    private SMEVMessage getRequest(QueryInformation query) throws SMEVProcessingException, SMEVRuntimeException {
        GetRequestRequest requestRequest = messageMapper.createGetRequestRequest(query);
        requestRequest.setCallerInformationSystemSignature(sign((requestRequest.getMessageTypeSelector())));
        if (loggingInput) {
            logMessage(requestRequest);
        }
        if (validator != null && validateInput) {
            validator.validate(requestRequest);
        }
        try {
            GeoMessage<GetRequestResponse> rs = geoTransport != null ?
                    geoTransport.getRequest(requestRequest) :
                    new GeoMessage<>(transport.getRequest(requestRequest));
            if (rs.content == null || rs.content.getRequestMessage() == null) {
                return null;
            }
            if (loggingOutput) {
                logMessage(rs.content);
            }
            if (validator != null && validateOutput) {
                logger.debug("Validating response message");
                validator.validate(rs.content);
            }
            if (validateSMEVSignature) {
                signer.validateSMEVSignature(
                        rs.content.getRequestMessage().getSMEVSignature().getAny(),
                        Marshaller.marshal(rs.content.getRequestMessage().getRequest()),
                        postValidationCallback);
            }
            SMEVMessage result = messageMapper.convertGetRequestResponse(rs.content);
            downloadAttachments(result);
            final SMEVMetadata metadata = result.getSMEVMetadata();
            final SMEVContext context = metadata.getSmevContext();
            metadata.setSmevContext(new SMEVContext(rs.idTransport, context));
            return result;
        } catch (Exception e) {
            logger.error("Ошибка получения запроса", e);
            throw exceptionMapper.convert(e);
        }
    }

    private void logMessage(Object element) {
        try {
            logger.info("Message: {}", Marshaller.convertElementToString(element));
        } catch (TransformerException | SMEVRuntimeException e) {
            logger.error("Ошибка логирования", e);
        }
    }


    private SMEVMessage getResponse(QueryInformation query) throws SMEVRuntimeException, SMEVProcessingException {
        GetResponseRequest responseRequest = messageMapper.createGetResponseRequest(query);
        responseRequest.setCallerInformationSystemSignature(sign((responseRequest.getMessageTypeSelector())));

        if (loggingInput) {
            logMessage(responseRequest);
        }
        if (validator != null && validateInput) {
            validator.validate(responseRequest);
        }
        try {
            GeoMessage<GetResponseResponse> rs = geoTransport != null ?
                    geoTransport.getResponse(responseRequest) :
                    new GeoMessage<>(transport.getResponse(responseRequest));
            if (rs.content == null || rs.content.getResponseMessage() == null) {
                return null;
            }
            if (loggingOutput) {
                logMessage(rs.content);
            }
            if (validator != null && validateOutput) {
                validator.validate(rs.content);
            }
            if (validateSMEVSignature) {
                signer.validateSMEVSignature(
                        rs.content.getResponseMessage().getSMEVSignature().getAny(),
                        Marshaller.marshal(rs.content.getResponseMessage().getResponse()),
                        postValidationCallback);
            }
            SMEVMessage result = messageMapper.convertGetResponseResponse(rs.content);
            downloadAttachments(result);
            final SMEVMetadata metadata = result.getSMEVMetadata();
            final SMEVContext context = metadata.getSmevContext();
            metadata.setSmevContext(new SMEVContext(rs.idTransport, context));
            return result;
        } catch (Exception e) {
            logger.error("Ошибка получения ответа", e);
            throw exceptionMapper.convert(e);
        }
    }

    private SMEVMessage getStatus() throws SMEVRuntimeException, SMEVProcessingException {
        GetStatusRequest statusRequest = messageMapper.createGetStatusRequest();
        statusRequest.setCallerInformationSystemSignature(sign((statusRequest.getTimestamp())));
        if (loggingInput) {
            logMessage(statusRequest);
        }
        if (validator != null && validateInput) {
            validator.validate(statusRequest);
        }
        try {
            GeoMessage<GetStatusResponse> rs = geoTransport != null ?
                    geoTransport.getStatus(statusRequest) :
                    new GeoMessage<>(transport.getStatus(statusRequest));
            if (rs.content == null || rs.content.getSmevAsyncProcessingMessage() == null) {
                return null;
            }
            if (loggingOutput) {
                logMessage(rs.content);
            }
            if (validator != null && validateOutput) {
                validator.validate(rs.content);
            }
            if (validateSMEVSignature) {
                signer.validateSMEVSignature(
                        rs.content.getSmevAsyncProcessingMessage().getSMEVSignature().getAny(),
                        Marshaller.marshal(rs.content.getSmevAsyncProcessingMessage().getAsyncProcessingStatusData()),
                        postValidationCallback);
            }
            final SMEVMessage message = messageMapper.convertStatusResponse(rs.content);
            final SMEVMetadata metadata = message.getSMEVMetadata();
            final SMEVContext context = metadata.getSmevContext();
            metadata.setSmevContext(new SMEVContext(rs.idTransport, context));
            return message;
        } catch (Exception e) {
            logger.error("Ошибка получения статуса", e);
            throw exceptionMapper.convert(e);
        }
    }

    /**
     * @param message the message
     * @return
     * @throws SMEVException
     */
    @Override
    public SMEVMetadata send(SMEVMessage message) throws SMEVException {
        if (!((message.getData().getStatus().equals(BusinessStatus.REQUEST)) || message.getData().getStatus().equals(BusinessStatus.RESPONSE))) {
            if (message.getData().getAttachments() != null && message.getData().getAttachments().size() > 0) {
                throw new SMEVRuntimeException("Аттачменты доступны только для ответов и запросов!");
            }
        }

        sendLargeAttachments(message.getData().getAttachments());

        if (message.getSMEVMetadata().getProcessingInformation() instanceof RequestInformation) {
            if (message.getData() instanceof AbstractResponseContent) {
                throw new SMEVRuntimeException("Несовместимые тип запроса и тип данных");  //TODO:
            }
            return sendRequest(message);
        }
        if (message.getSMEVMetadata().getProcessingInformation() instanceof ResponseInformation) {
            if (message.getData() instanceof AbstractRequestContent) {
                throw new SMEVRuntimeException("Несовместимые тип запроса и тип данных"); //TODO:
            }
            return sendResponse(message);
        }
        return null;
    }

    private void sendLargeAttachments(List<SMEVAttachment> attachments) throws SMEVRuntimeException {
        if (attachments == null || largeAttachmentTransport == null) {
            logger.debug("LargeAttachments will not be processed to deliver");
            return;
        }
        for (SMEVAttachment attachment : attachments) {
            if (attachment instanceof LargeAttachment) {
                largeAttachmentTransport.upload((LargeAttachment) attachment);
            }
        }
    }

    private SMEVMetadata sendRequest(SMEVMessage message) throws SMEVRuntimeException, SMEVProcessingException {
        SendRequestRequest requestRequest = messageMapper.createSendRequestRequest(message);
        requestRequest.setCallerInformationSystemSignature(sign((requestRequest.getSenderProvidedRequestData())));
        if (loggingInput) {
            logMessage(requestRequest);
        }
        if (validator != null && validateInput) {
            validator.validate(requestRequest);
        }
        try {
            SMEVMetadata response = new SMEVMetadata(
                    message.getSMEVMetadata().getMessageIdentity(),
                    message.getSMEVMetadata().getProcessingInformation(),
                    message.getSMEVMetadata().getExtendedProperties());
            GeoMessage<SendRequestResponse> rs = geoTransport != null ?
                    geoTransport.sendRequest(requestRequest) :
                    new GeoMessage<>(transport.sendRequest(requestRequest));

            if (loggingOutput) {
                logMessage(rs.content);
            }
            if (validator != null && validateOutput) {
                validator.validate(rs.content);
            }
            if (validateSMEVSignature) {
                signer.validateSMEVSignature(rs.content.getSMEVSignature().getAny(),
                        Marshaller.marshal(rs.content.getMessageMetadata()),
                        postValidationCallback);
            }
            final SMEVContext context = messageMapper.convertSendRequestResponse(rs.content).getSmevContext();
            response.setSmevContext(new SMEVContext(rs.idTransport, context));
            return response;
        } catch (Exception e) {
            logger.error("Ошибка отправки запроса", e);
            throw exceptionMapper.convert(e);
        }
    }

    /**
     * @param message
     * @return
     * @throws SMEVRuntimeException
     * @throws SMEVProcessingException
     */
    private SMEVMetadata sendResponse(SMEVMessage message) throws SMEVRuntimeException, SMEVProcessingException {
        SendResponseRequest responseRequest = messageMapper.createSendResponseRequest(message);
        responseRequest.setCallerInformationSystemSignature(sign((responseRequest.getSenderProvidedResponseData())));
        if (loggingInput) {
            logMessage(responseRequest);
        }
        if (validator != null && validateInput) {
            validator.validate(responseRequest);
        }
        try {
            SMEVMetadata response = new SMEVMetadata(
                    message.getSMEVMetadata().getMessageIdentity(),
                    message.getSMEVMetadata().getProcessingInformation(),
                    message.getSMEVMetadata().getExtendedProperties());
            GeoMessage<SendResponseResponse> rs = geoTransport != null ?
                    geoTransport.sendResponse(responseRequest) :
                    new GeoMessage<>(transport.sendResponse(responseRequest));
            if (loggingOutput) {
                logMessage(rs.content);
            }
            if (validator != null && validateOutput) {
                validator.validate(rs.content);
            }
            if (validateSMEVSignature) {
                signer.validateSMEVSignature(
                        rs.content.getSMEVSignature().getAny(),
                        Marshaller.marshal(rs.content.getMessageMetadata()),
                        postValidationCallback);
            }
            final SMEVContext context = messageMapper.convertSendResponseResponse(rs.content).getSmevContext();
            response.setSmevContext(new SMEVContext(rs.idTransport, context));
            return response;
        } catch (Exception e) {
            logger.error("Ошибка отправки ответа", e);
            throw exceptionMapper.convert(e);
        }
    }

    @Override
    public void ack(SMEVMetadata metadata, boolean accepted) throws SMEVException {
        try {
            AckRequest request = messageMapper.createAckRequest(metadata.getMessageIdentity().getMessageId(), accepted);
            request.setCallerInformationSystemSignature(sign((request.getAckTargetMessage())));
            if (loggingInput) {
                logMessage(request);
            }
            if (validator != null && validateInput) {
                validator.validate(request);
            }
            if (geoTransport != null) {
                String idTransport = null;
                final SMEVContext context = metadata.getSmevContext();
                if (context != null) {
                    idTransport = context.getIdTransport();
                }
                geoTransport.ack(new GeoMessage<>(idTransport, request));
            } else if (transport != null) {
                transport.ack(request);
            } else {
                throw new SMEVException("Ни один транспорт не определен");
            }
        } catch (Exception e) {
            logger.error("Ошибка отправки подтверждения", e);
            throw exceptionMapper.convert(e);
        }
    }

    @Override
    public Map<String, Long> getIncomingQueueStatistics(String nodeId) throws SMEVException {
      throw new UnsupportedOperationException();
    }

    private XMLDSigSignatureType sign(Object object) throws SMEVRuntimeException, SMEVSignatureException {
        XMLDSigSignatureType signatureWrapper = new XMLDSigSignatureType();
        signatureWrapper.setAny(signer.signXMLDSigDetached(Marshaller.marshal(object), null));
        return signatureWrapper;
    }

    private void downloadAttachments(SMEVMessage message) throws Exception {
        if ((localStorage == null) || (message == null) || (message.getData() == null) ||
                (message.getData().getAttachments() == null)) {
            return;
        }
        for (SMEVAttachment attachment : message.getData().getAttachments()) {
            if (attachment instanceof MTOMAttachment) {   //add folder
                saveFile(localStorage, (MTOMAttachment) attachment);
            }
            if (attachment instanceof LargeAttachment) {   //add folder
                saveLargeFile(localStorage, (LargeAttachment) attachment);
            }
        }
    }

    private void saveFile(String localstorage, MTOMAttachment attachment) throws Exception {
        File file = new File(Paths.get(localstorage, attachment.getAttachmentId()).toUri());
        try {
            Files.copy(attachment.getContent().getInputStream(), file.toPath());
        } finally {
            try {
                attachment.getContent().getInputStream().close();
            } catch (IOException e) {
                logger.debug("Ошибка закрытия входного аттачмента", e);
            }
        }
    }

    private void saveLargeFile(String localstorage, LargeAttachment attachment) throws Exception {
        if (largeAttachmentTransport == null) {
            return;
        }
        File file = new File(Paths.get(localstorage, attachment.getContentRef().getRawPath()).toUri());
        OutputStream output = null;
        try {
            if (!file.exists()) {
                file.createNewFile();
            }
            output = new FileOutputStream(file);
            largeAttachmentTransport.download(attachment, output);
            output.flush();
        } finally {
            if (output != null) {
                try {
                    output.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }
}
