package ru.voskhod.smev.client.api.transport.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.voskhod.smev.client.api.services.transport.GeoMessageTransport;
import ru.voskhod.smev.client.api.services.transport.configuration.GeoMessageTransportConfiguration;
import ru.voskhod.smev.client.api.services.transport.configuration.TransportEndpointConfiguration;
import ru.voskhod.smev.client.api.transport.Constants;
import ru.voskhod.smev.client.api.types.message.transport.GeoMessage;
import ru.voskhod.smev.server.autogenerated.SMEVMessageExchangePortType;
import ru.voskhod.smev.server.autogenerated.SMEVMessageExchangeService;
import ru.voskhod.smev.server.autogenerated.types.basic.Void;
import ru.voskhod.smev.server.autogenerated.types.*;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.soap.SOAPBinding;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * The type Message transport.
 */
public final class GeoMessageTransportImpl implements GeoMessageTransport<SendRequestRequest, SendRequestResponse,
        SendResponseRequest, SendResponseResponse,
        GetRequestRequest, GetRequestResponse,
        GetResponseRequest, GetResponseResponse,
        GetStatusRequest, GetStatusResponse,
        AckRequest, Void,
        Object, Object> {

    private final Map<String, Transport> transports = new HashMap<>();
    private final List<String> transportIds = new ArrayList<>();
    private static final Logger logger = LoggerFactory.getLogger(GeoMessageTransportImpl.class);

    private int current = -1;

    /**
     * Instantiates a new Message transport.
     *
     * @param config the config
     * @throws Exception the exception
     */
    public <T extends GeoMessageTransportConfiguration> GeoMessageTransportImpl(T config) throws Exception {
        for (TransportEndpointConfiguration conf : config.getEndpointConfigurations()) {
            final String url = conf.getUrl();
            final String id = conf.getId();
            final TransportProvider provider = new TransportProvider(
                    url,
                    Constants.EXCHANGE_MESSAGE_SERVICE,
                    conf.getTimeout()
            );
            transports.put(id, new Transport(id, url, provider, conf.getRetries()));
            transportIds.add(id);
        }
    }


    @Override
    public GeoMessage<SendRequestResponse> sendRequest(final SendRequestRequest parameters) throws Exception {
        return callTransport(new SuccessPortCall<SendRequestResponse>() {
            @Override
            public SendRequestResponse call(SMEVMessageExchangePortType port) throws Exception {
                return port.sendRequest(parameters);
            }
        }, "Ни одному транспорту не удалось отправить запрос");
    }

    @Override
    public GeoMessage<SendResponseResponse> sendResponse(final SendResponseRequest parameters) throws Exception {
        return callTransport(new SuccessPortCall<SendResponseResponse>() {
            @Override
            public SendResponseResponse call(SMEVMessageExchangePortType port) throws Exception {
                return port.sendResponse(parameters);
            }
        }, "Ни одному транспорту не удалось отправить ответ");
    }

    @Override
    public GeoMessage<GetRequestResponse> getRequest(final GetRequestRequest parameters) throws Exception {
        return callTransport(new PortCall<GetRequestResponse>() {
            @Override
            public GetRequestResponse call(SMEVMessageExchangePortType port)
                    throws Exception {
                return port.getRequest(parameters);
            }

            @Override
            public boolean isSuccess(GetRequestResponse response) {
                return isSuccessRequest(response);
            }
        }, null);
    }

    @Override
    public GeoMessage<GetStatusResponse> getStatus(final GetStatusRequest parameters) throws Exception {
        return callTransport(new SuccessPortCall<GetStatusResponse>() {
            @Override
            public GetStatusResponse call(SMEVMessageExchangePortType port)
                    throws Exception {
                return port.getStatus(parameters);
            }
        }, "Ни одному транспорту не удалось получить статус");
    }

    @Override
    public GeoMessage<GetResponseResponse> getResponse(final GetResponseRequest parameters) throws Exception {
        return callTransport(new PortCall<GetResponseResponse>() {
            @Override
            public GetResponseResponse call(SMEVMessageExchangePortType port)
                    throws Exception {
                return port.getResponse(parameters);
            }

            @Override
            public boolean isSuccess(GetResponseResponse response) {
                return isSuccessResponse(response);
            }
        }, null);
    }

    private <ResponseCall> GeoMessage<ResponseCall> callTransport(PortCall<ResponseCall> call, String error) {
        Transport transport;
        do {
            transport = currentTransport();
            for (int i = 0; i < transport.retries; ++i) {
                try {
                    final ResponseCall result = call.call(transport.provider.port());
                    if (call.isSuccess(result)) {
                        return new GeoMessage<>(transport.id, result);
                    } else {
                        break;
                    }
                } catch (Exception ex) {
                    logger.error("Exception occured while trying to call transport, the exception is: ", ex);
                }
            }
        } while (nextNotLast(transport));
        if (error != null) {
            throw new RuntimeException(error);
        }
        return new GeoMessage<>(null, null);
    }

    private interface PortCall<ResponseCall> {

        ResponseCall call(SMEVMessageExchangePortType port) throws Exception;

        boolean isSuccess(ResponseCall response);
    }

    private abstract class SuccessPortCall<ResponseCall> implements PortCall<ResponseCall> {
        @Override
        public boolean isSuccess(ResponseCall response) {
            return true;
        }
    }

    private static SMEVMessageExchangePortType createPort(String url, QName serviceQName, int timeout)
            throws Exception {

        SMEVMessageExchangeService service = new SMEVMessageExchangeService(new URL(url), serviceQName);

        SMEVMessageExchangePortType port = service.getSMEVMessageExchangeEndpoint();
        ((BindingProvider) port).getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", timeout);
        ((BindingProvider) port).getRequestContext().put("com.sun.xml.internal.ws.request.timeout", timeout);
        ((BindingProvider) port).getRequestContext().put("com.sun.xml.ws.request.timeout", timeout);
        ((BindingProvider) port).getRequestContext().put("com.sun.xml.ws.connect.timeout", timeout);

        SOAPBinding binding = (SOAPBinding) ((BindingProvider) port).getBinding();
        binding.setMTOMEnabled(true);
        return port;
    }

    protected final Transport currentTransport() {
        normalize();
        final String id = transportIds.get(current);
        return transports.get(id);
    }

    /**
     * Round-robin
     */
    protected final void next() {
        ++current;
        normalize();
    }

    protected final boolean nextNotLast(Transport transport) {
        if (isLast(transport))
            return false;
        next();
        return true;
    }

    protected final void normalize() {
        if (current >= transportIds.size() || current < 0)
            current = 0;
    }

    protected final boolean isLast(Transport transport) {
        int i = transportIds.indexOf(transport.id);
        return transportIds.size() - 1 == i;
    }

    protected final Transport getTransport(String id) {
        if (id == null)
            return null;
        return transports.get(id);
    }

    protected static boolean isSuccessResponse(GetResponseResponse response) {
        return response.getResponseMessage() != null;
    }

    protected static boolean isSuccessRequest(GetRequestResponse response) {
        return response.getRequestMessage() != null;
    }

    @Override
    public GeoMessage<Void> ack(GeoMessage<AckRequest> parameters)
            throws Exception {
        final Transport transport = getTransport(parameters.idTransport);
        if (transport == null) {
            return tryAck(parameters.content);
        }
        final Void result = transport.provider.port()
                .ack(parameters.content);
        return new GeoMessage<>(transport.id, result);
    }

    private GeoMessage<Void> tryAck(AckRequest request) throws Exception {
        for (Transport transport : transports.values()) {
            try {
                final Void result = transport.provider.port().ack(request);
                return new GeoMessage<>(transport.id, result);
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        throw new RuntimeException("Ни один транспорт не содержит подтверждаемое сообщение");
    }

    static final class Transport {
        final String id;
        final String url;
        final TransportProvider provider;
        final int retries;

        private Transport(String id, String url, TransportProvider provider, int retries) {
            this.id = id;
            this.url = url;
            this.provider = provider;
            this.retries = retries;
        }
    }
}
