package ru.voskhod.smev.client.api.services.validation.impl;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import ru.voskhod.smev.client.api.services.validation.Validator;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.exception.processing.SMEVMessageContentException;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;


/**
 * The type Smev schema validator.
 */
public class SMEVSchemaValidatorImpl implements Validator<Object> {

    /**
     * The constant schema.
     */
    final static Schema schema = init();

    /**
     * Init schema.
     *
     * @return the schema
     */
    static Schema init() {
        SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        try {
            return factory.newSchema(new Source[]{    //hmmm  -maybe multiple loads
                    new StreamSource(SMEVSchemaValidatorImpl.class.getResourceAsStream("/smev-message-exchange-basic-1.2.xsd")),
                    new StreamSource(SMEVSchemaValidatorImpl.class.getResourceAsStream("/smev-message-exchange-faults-1.2.xsd")),
                    new StreamSource(SMEVSchemaValidatorImpl.class.getResourceAsStream("/smev-message-exchange-types-1.2.xsd"))
            });
        } catch (Exception e) {
            throw new IllegalStateException("Unable to load xml scheme", e);
        }
    }


    @Override
    public void validate(Object toValidate) throws SMEVRuntimeException, SMEVMessageContentException.InvalidContentException {
        List<XMLValidationError> validationErrorList;
        try {
            validationErrorList = validateContent(SMEVMarshaller.marshal(toValidate));
        } catch (SAXException | IOException e) {
            throw new SMEVRuntimeException(e);
        }

        if (validationErrorList != null && validationErrorList.size() > 0) {

            SMEVMessageContentException.InvalidContentException.InvalidContent invalidContent = new SMEVMessageContentException.InvalidContentException.InvalidContent();
            for (XMLValidationError src : validationErrorList) {
                SMEVMessageContentException.InvalidContentException.InvalidContent.ValidationError dst = new SMEVMessageContentException.InvalidContentException.InvalidContent.ValidationError();
                dst.setErrorPosition(src.getRow());
                dst.setValue(src.getMessage());
                invalidContent.getValidationError().add(dst);
            }
            throw new SMEVMessageContentException.InvalidContentException("Invalid schema", invalidContent);
        }
    }

    private javax.xml.validation.Validator getValidator() throws SAXException {

        if (schema == null)
            throw new SAXException("no schema defined");

        return schema.newValidator();

    }

    private List<XMLValidationError> validateContent(Element argContent) throws SAXException, IOException {
        javax.xml.validation.Validator validator = getValidator();
        // Валидация по XSD.
        Source validationSource = new DOMSource(argContent);
        ValidationErrorHandler errorHandler = new ValidationErrorHandler();
        validator.setErrorHandler(errorHandler);

        validator.validate(validationSource);


        List<XMLValidationError> result = errorHandler.result;

        return result;
    }

    /**
     * The type Validation error handler.
     */
    static class ValidationErrorHandler implements ErrorHandler {

        /**
         * The Result.
         */
        public final List<XMLValidationError> result = new LinkedList<>();

        @Override
        public void error(SAXParseException arg) {
            XMLValidationError error = new XMLValidationError();
            error.setRow(arg.getLineNumber());
            error.setColumn(arg.getColumnNumber());
            error.setMessage(arg.getMessage());
            result.add(error);
        }

        @Override
        public void fatalError(SAXParseException arg) {
            XMLValidationError error = new XMLValidationError();
            error.setRow(arg.getLineNumber());
            error.setColumn(arg.getColumnNumber());
            error.setMessage(arg.getMessage());
            result.add(error);
        }

        @Override
        public void warning(SAXParseException arg) {
            XMLValidationError error = new XMLValidationError();
            error.setRow(arg.getLineNumber());
            error.setColumn(arg.getColumnNumber());
            error.setMessage(arg.getMessage());
            result.add(error);
        }
    }

    /**
     * The type Xml validation error.
     */
    static class XMLValidationError {

        private int row;
        private int column;
        private String message;

        /**
         * Gets row.
         *
         * @return the row
         */
        public int getRow() {
            return row;
        }

        /**
         * Sets row.
         *
         * @param row the row
         */
        public void setRow(int row) {
            this.row = row;
        }

        /**
         * Gets column.
         *
         * @return the column
         */
        public int getColumn() {
            return column;
        }

        /**
         * Sets column.
         *
         * @param column the column
         */
        public void setColumn(int column) {
            this.column = column;
        }

        /**
         * Gets message.
         *
         * @return the message
         */
        public String getMessage() {
            return message;
        }

        /**
         * Sets message.
         *
         * @param message the message
         */
        public void setMessage(String message) {
            this.message = message;
        }
    }


    /**
     * The type Smev marshaller.
     */
    static class SMEVMarshaller {
        /**
         * The constant SERVER_API_PACKAGE.
         */
        public static final String SERVER_API_PACKAGE = "ru.voskhod.smev.server.autogenerated.types";
        private static final ThreadLocal<JAXBContext> jaxbContext = new ThreadLocal<>();

        private static void init() {
            if (jaxbContext.get() == null) {
                try {
                    jaxbContext.set(JAXBContext.newInstance(SERVER_API_PACKAGE));
                } catch (JAXBException e) {
                    throw new RuntimeException(e);
                }
            }
        }

        /**
         * Marshal element.
         *
         * @param obj the obj
         * @return the element
         * @throws SMEVRuntimeException the smev runtime exception
         */
        public static Element marshal(Object obj) throws SMEVRuntimeException {
            if (jaxbContext.get() == null) { //bad
                init();
            }
            DOMResult domResult = new DOMResult();

            try {
                Marshaller marshaller = jaxbContext.get().createMarshaller();
                //marshaller.setSchema(schema);

                marshaller.marshal(obj, domResult);
            } catch (JAXBException e) {
                throw new SMEVRuntimeException(e.getMessage(), e);
            }

            return ((Document) domResult.getNode()).getDocumentElement();
        }

        /**
         * Unmarshal object.
         *
         * @param element the element
         * @return the object
         * @throws SMEVRuntimeException the smev runtime exception
         */
        public static Object unmarshal(Element element) throws SMEVRuntimeException {
            if (jaxbContext.get() == null) {
                init();
            }
            try {
                return jaxbContext.get().createUnmarshaller().unmarshal(element);
            } catch (JAXBException e) {
                throw new SMEVRuntimeException(e.getMessage(), e);
            }
        }
    }
}


