package ru.voskhod.smev.client.toolset.commons;

import org.apache.commons.cli.*;
import org.apache.commons.io.IOUtils;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * Базовый класс для утилит
 */
public abstract class AbstractTool {

    /**
     * Параметр командной строки: операция для выполнения
     */
    public static final String PARAM_CMD = "cmd";

    /**
     * Параметр командной строки: входной файл
     */
    public static final String PARAM_INPUT_FILE = "if";

    /**
     * Параметр командной строки: файл результата
     */
    public static final String PARAM_OUTPUT_FILE = "of";

    /**
     * Параметр командной строки: файл с аргументами командной строки
     * Аргументы можно указывать как непосредственно в самой командной строке, так и в этом файле.
     */
    public static final String PARAM_ARGS_FILE = "argsFile";

    /**
     * Параметр командной строки: идентификатор сообщения
     */
    public static final String PARAM_MESSAGE_ID = "msgId";

    /**
     * Параметр командной строки: идентификатор изначального сообщения
     */
    public static final String PARAM_REFERENCE_MESSAGE_ID = "refMsgId";

    /**
     * Параметр командной строки: адрес ответа
     */
    public static final String PARAM_REPLY_TO = "replyTo";

    /**
     * Параметр командной строки: идентификатор транзакции
     */
    public static final String PARAM_TRANSACTION_ID = "tid";

    /**
     * Описание параметров командной строки
     */
    private Options options;

    /**
     * Командная строка
     */
    protected CommandLine commandLine;

    /**
     * Конструктор
     *
     * @param args параметры командной строки
     */
    public AbstractTool(String[] args) {
        //заполняем описание командной строки
        options = new Options();
        Option paramFile = Option.builder(PARAM_ARGS_FILE).hasArg().argName("имя файла").desc("Файл с аргументами командной строки. Аргументы можно указывать как непосредственно в самой командной строке, так и в этом файле.").build();
        options.addOption(paramFile);
        populateOptions(options);

        //добавляем аргументы из файла (если указан)
        List<String> preparedArgs = new ArrayList<>();
        String argsFileParam = "-" + PARAM_ARGS_FILE;
        for (int i = 0; i < args.length; ++i) {
            if (argsFileParam.equals(args[i])) {
                if (++i < args.length) {
                    addArgsFromFile(args[i], preparedArgs);
                }
            } else {
                preparedArgs.add(args[i]);
            }
        }

        //парсим командную строку
        try {
            commandLine = new DefaultParser().parse(options, preparedArgs.toArray(new String[0]));
        } catch (ParseException ex) {
            printUsage();
            throw new IllegalStateException("Неверные аргументы командной строки", ex);
        }
    }

    /**
     * Добавляет параметры из указанного файла в указанный список
     *
     * @param file путь к файлу с аргументами
     * @param args список аргументов
     */
    private void addArgsFromFile(String file, List<String> args) {
        InputStream fileIS = null;
        try {
            fileIS = new FileInputStream(file);
            Properties properties = new Properties();
            properties.load(fileIS);
            for (String name : properties.stringPropertyNames()) {
                args.add("-" + name);
                args.add(properties.getProperty(name));
            }
        } catch (Exception ex) {
            printUsage();
            throw new IllegalStateException("Ошибка чтения файла с аргументами командной строки", ex);
        } finally {
            IOUtils.closeQuietly(fileIS);
        }
    }

    /**
     * Добавляет специфические для данной утилиты параметры
     *
     * @param options описание параметров командной строки
     */
    protected abstract void populateOptions(Options options);

    /**
     * Проверяет наличие указанных аргументов в командной строке
     *
     * @param args параметры, наличие которых необходимо проверить
     * @throws ParseException если какого-то из параметров нет
     */
    protected void requireArgs(String... args) throws ParseException {
        for (String paramName : args) {
            if (!commandLine.hasOption(paramName)) {
                throw new ParseException("Необходимо указать параметр: " + paramName);
            }
        }
    }

    /**
     * Отображает формат вызова и параметры командной строки утилиты
     */
    protected void printUsage() {
        HelpFormatter formatter = new HelpFormatter();
        formatter.setWidth(120);
        formatter.setSyntaxPrefix("");
        formatter.printHelp("Параметры:", null, options, getHelpFooter());
    }

    /**
     * Возвращает нижнюю подпись к помощи по командной строке
     *
     * @return подпись к помощи по командной строке
     */
    protected String getHelpFooter() {
        return null;
    }

    /**
     * Парсит XML файл и возвращает корневой элемент
     *
     * @param file файл с xml документом
     * @return корневой элемент документа
     * @throws IOException                  в случае ошибки чтения входного файла
     * @throws ParserConfigurationException в случае ошибки создания xml парсера
     * @throws SAXException                 в случае невалидного xml кода
     */
    protected static Element fileToElement(File file) throws IOException, ParserConfigurationException, SAXException {
        if (file == null) {
            return null;
        }
        try (InputStream inputStream = new FileInputStream(file)) {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
            documentBuilderFactory.setNamespaceAware(true);
            documentBuilderFactory.setCoalescing(true);
            documentBuilderFactory.setIgnoringElementContentWhitespace(true);
            return documentBuilderFactory.newDocumentBuilder().parse(inputStream).getDocumentElement();
        }
    }

    /**
     * Возворащает входной файл, указанный в командной строке
     * Также проверяет существование файла
     *
     * @param commandLine параметры командной строки
     * @return входной файл
     */
    protected static File getInputFile(CommandLine commandLine) {
        File file = new File(commandLine.getOptionValue(PARAM_INPUT_FILE));
        if (file.exists() && file.isFile()) {
            return file;
        }
        throw new IllegalArgumentException("Входной файл не существует");
    }

    /**
     * Возвращает аргумент командной строки в виде квалифицированного имени
     *
     * @param paramName имя аргумента командной строки
     * @return распарсенное имя
     * @throws ParseException если не удалось распарсить параметр
     */
    protected QName getQNameOptionValue(String paramName) throws ParseException {
        String str = commandLine.getOptionValue(paramName);
        if (str == null) {
            return null;
        }
        try {
            return QName.valueOf(str);
        } catch (Exception ex) {
            throw new ParseException("Неверный параметр " + paramName);
        }
    }

    /**
     * Парсит список файлов, переданный в виде параметра командной строки
     *
     * @param paramName имя аргумента командной строки
     * @return список файлов
     * @throws ParseException если указан несуществующий файл
     */
    protected List<File> getFileListOptionValue(String paramName) throws ParseException {
        String listString = commandLine.getOptionValue(paramName);
        if (listString == null) {
            return Collections.emptyList();
        }
        List<File> result = new ArrayList<>();
        for (String str : listString.split(";")) {
            File f = new File(str);
            if (!f.exists()) {
                throw new ParseException("Файл не существует: " + str);
            }
            result.add(f);
        }
        return result;
    }

    /**
     * Записывает указанные данные в выходной файл или консоль
     *
     * @param element данные для записи
     * @param suffix  суффикс для имени файла, null - без суффикса
     * @throws IOException в случае ошибки ввода-вывода
     */
    protected void writeOutput(Element element, String suffix) throws IOException {
        if (element == null) {
            return;
        }
        try {
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            StreamResult result = new StreamResult(new StringWriter());
            DOMSource source = new DOMSource(element);
            transformer.transform(source, result);
            writeOutput(result.getWriter().toString(), suffix);
        } catch (TransformerException te) {
            throw new IOException("Не удалось сконвертировать xml-элемент в строку", te);
        }
    }

    /**
     * Записывает указанные данные в выходной файл
     *
     * @param string данные для записи
     * @param suffix суффикс для имени файла, null - без суффикса
     * @throws IOException если запись не удалась
     */
    protected void writeOutput(String string, String suffix) throws IOException {
        if (string == null) {
            return;
        }
        writeOutput(string.getBytes(), suffix);
    }

    /**
     * Записывает набор свойств и их значений в выходной файл или консоль
     *
     * @param props  набор свойств и их значений
     * @param suffix суффикс для имени файла, null - без суффикса
     * @throws IOException в случае ошибки ввода-вывода
     */
    protected void writeOutput(Properties props, String suffix) throws IOException {
        if ((props == null) || props.isEmpty()) {
            return;
        }
        StringWriter sw = new StringWriter();
        props.store(sw, null);
        writeOutput(sw.toString(), suffix);
    }

    /**
     * Записывает указанные данные в выходной файл или консоль
     *
     * @param data   данные для записи
     * @param suffix суффикс для имени файла, null - без суффикса
     * @throws IOException в случае ошибки ввода-вывода
     */
    protected void writeOutput(byte[] data, String suffix) throws IOException {
        if (data == null) {
            return;
        }
        OutputStream os = null;
        boolean needClosing = false;
        try {
            if (commandLine.hasOption(PARAM_OUTPUT_FILE)) {
                os = new FileOutputStream(commandLine.getOptionValue(PARAM_OUTPUT_FILE) + (suffix == null ? "" : suffix));
                needClosing = true;
            } else {
                os = System.out;
            }
            IOUtils.write(data, os);
        } finally {
            if (needClosing) {
                IOUtils.closeQuietly(os);
            }
        }
    }

    /**
     * Возвращает значение целочисленного аргумента командной строки
     *
     * @param argName имя аргумента
     * @return значение аргумента
     * @throws ParseException если указанный аргумент не является целым числом
     */
    protected int getIntArg(String argName) throws ParseException {
        try {
            return Integer.parseInt(commandLine.getOptionValue(argName));
        } catch (Exception ex) {
            throw new ParseException(String.format("Значение параметра %s не является числом", argName));
        }
    }

    /**
     * Возвращает аргумент командной строки в виде даты
     *
     * @param argName имя аргумента
     * @return значение аргумента
     * @throws ParseException если указанный аргумент не является датой
     */
    protected Date getDateArg(String argName) throws ParseException {
        if (!commandLine.hasOption(argName)) {
            return null;
        }
        try {
            return new SimpleDateFormat("dd.MM.yyyy HH:mm:ss").parse(commandLine.getOptionValue(argName));
        } catch (Exception ex) {
            throw new ParseException(String.format("Значение параметра %s не является датой", argName));
        }
    }

    /**
     * Возвращает набор свойств из указанного файла
     *
     * @param argName имя аргумента, указывающего на файл со свойствами
     * @return набор свойств
     * @throws ParseException если возникла ошибка при парсинге файла свойств
     */
    protected Map<String, String> getMapArg(String argName) throws ParseException {
        String fileName = commandLine.getOptionValue(argName);
        if (fileName == null) {
            return null;
        }
        FileInputStream fis = null;
        try {
            Properties prop = new Properties();
            fis = new FileInputStream(fileName);
            prop.load(fis);
            Map<String, String> result = new HashMap<>();
            for (String propName : prop.stringPropertyNames()) {
                result.put(propName, prop.getProperty(propName));
            }
            return result;
        } catch (Exception ex) {
            throw new ParseException(String.format("Параметр %s должен указывать на файл с параметрами", argName));
        } finally {
            IOUtils.closeQuietly(fis);
        }
    }
}

