package ru.voskhod.smev.client.toolset.identification;

import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.identification.IdentityService;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.toolset.commons.AbstractTool;
import ru.voskhod.smev.client.toolset.commons.ToolConfiguration;

/**
 * Утилита для использования {@link ru.voskhod.smev.client.api.services.identification.IdentityService генератора идентификаторов}
 */
public class Main extends AbstractTool {

    /**
     * Операция генерации идентификатора сообщения
     */
    public static final String CMD_GENERATE_MESSAGE_UUID = "msgUuid";

    /**
     * Операция генерации идентификатора вложения по номеру и MIME-типу
     */
    public static final String CMD_GENERATE_ATTACHMENT_ID = "attachId";

    /**
     * Операция генерации уникального идентификатора вложения
     */
    public static final String CMD_GENERATE_ATTACHMENT_UUID = "attachUuid";

    /**
     * Параметр командной строки: MIME-тип
     */
    public static final String PARAM_MIME_TYPE = "mime";

    /**
     * Параметр командной строки: номер вложения
     */
    public static final String PARAM_ATTACHMENT_NUMBER = "num";

    /**
     * Логгер
     */
    private static final Logger LOG = LoggerFactory.getLogger(Main.class);

    /**
     * Генератор идентификаторов
     */
    private IdentityService identityService;

    /**
     * Запускает утилиту
     *
     * @param args параметры командной строки
     */
    public static void main(String[] args) {
        Main tool = null;
        try {
            tool = new Main(args);
            tool.execute();
        } catch (ParseException pe) {
            if (tool != null) {
                tool.printUsage();
            }
            LOG.error("Ошибка аргументов командной строки", pe);
            System.exit(1);
        } catch (Exception ex) {
            LOG.error("Ошибка утилиты", ex);
            System.exit(1);
        }
    }

    /**
     * Создает экземпляр утилиты с указанными параметрам
     *
     * @param args параметры командной строки
     */
    private Main(String[] args) throws SMEVRuntimeException {
        super(args);
        identityService = Factory.getIdentityGeneratorInstance(ToolConfiguration.getDefault().getIdentityConfiguration());
    }

    /**
     * Исполняет команду, указанную в командной строке
     *
     * @throws Exception в случа возникновения ошибки
     */
    private void execute() throws Exception {
        String cmd = commandLine.getOptionValue(PARAM_CMD);
        if (cmd.equals(CMD_GENERATE_MESSAGE_UUID)) {
            writeOutput(identityService.generateUUID(), null);
        } else if (cmd.equals(CMD_GENERATE_ATTACHMENT_ID)) {
            requireArgs(PARAM_MIME_TYPE, PARAM_ATTACHMENT_NUMBER);
            writeOutput(identityService.generateAttachmentId(commandLine.getOptionValue(PARAM_MIME_TYPE),
                    getIntArg(PARAM_ATTACHMENT_NUMBER)), null);
        } else if (cmd.equals(CMD_GENERATE_ATTACHMENT_UUID)) {
            writeOutput(identityService.generateAttachmentUUID().toString(), null);
        } else {
            throw new ParseException("Неизвестная операция: " + cmd);
        }
    }

    /**
     * Добавляет специфические для данной утилиты параметры
     *
     * @param options описание параметров командной строки
     */
    @Override
    protected void populateOptions(Options options) {
        //команды
        String cmdDesc = String.format("Выполняемая операция. Возможные значения: %s - сгенерировать ID сообщения, " +
                        "%s - сгенерировать ID вложения по номеру и MIME-типу, " +
                        "%s - сгенерировать уникальный идентификатор вложения",
                CMD_GENERATE_MESSAGE_UUID, CMD_GENERATE_ATTACHMENT_ID, CMD_GENERATE_ATTACHMENT_UUID);
        Option cmd = Option.builder(PARAM_CMD).hasArg().required().argName("тип операции").desc(cmdDesc).build();
        options.addOption(cmd);
        //параметры
        Option mimeType = Option.builder(PARAM_MIME_TYPE).hasArg().argName("MIME-тип").desc("MIME-тип файла-вложения").build();
        options.addOption(mimeType);
        Option attachmentNumber = Option.builder(PARAM_ATTACHMENT_NUMBER).hasArg().argName("номер").desc("Номер файла-вложения по порядку.").build();
        options.addOption(attachmentNumber);
        Option outputFile = Option.builder(PARAM_OUTPUT_FILE).hasArg().argName("имя файла").desc("Путь к выходному файлу").build();
        options.addOption(outputFile);
    }
}
