package ru.voskhod.smev.client.toolset.message;

import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.template.Template;
import ru.voskhod.smev.client.api.services.template.configuration.GeoTemplateConfiguration;
import ru.voskhod.smev.client.api.types.exception.SMEVException;
import ru.voskhod.smev.client.api.types.exception.SMEVProcessingException;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.exception.processing.*;
import ru.voskhod.smev.client.api.types.message.SMEVMessage;
import ru.voskhod.smev.client.api.types.message.attachment.LargeAttachment;
import ru.voskhod.smev.client.api.types.message.attachment.MTOMAttachment;
import ru.voskhod.smev.client.api.types.message.attachment.SMEVAttachment;
import ru.voskhod.smev.client.api.types.message.business.data.response.RejectResponseContent;
import ru.voskhod.smev.client.api.types.message.business.data.response.StatusResponseContent;
import ru.voskhod.smev.client.api.types.message.business.enums.BusinessStatus;
import ru.voskhod.smev.client.api.types.message.system.SMEVContext;
import ru.voskhod.smev.client.api.types.message.system.SMEVMetadata;
import ru.voskhod.smev.client.toolset.commons.AbstractTool;
import ru.voskhod.smev.client.toolset.commons.ToolConfiguration;

import javax.xml.namespace.QName;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 * Утилита отправки сообщений СМЭВ
 */
public class Main extends AbstractTool {

    /**
     * Логгер
     */
    private static final Logger LOG = LoggerFactory.getLogger(Main.class);

    /**
     * Команда отправки запроса
     */
    public static final String CMD_SEND_REQUEST = "sendReq";

    /**
     * Команда получения запроса
     */
    public static final String CMD_GET_REQUEST = "getReq";

    /**
     * Команда отправки ответа
     */
    public static final String CMD_SEND_RESPONSE = "sendResp";

    /**
     * Команда получения ответа
     */
    public static final String CMD_GET_RESPONSE = "getResp";

    /**
     * Команда подтверждения получения сообщения
     */
    public static final String CMD_ACK = "ack";

    /**
     * Команда отправки ответа с отказом
     */
    public static final String CMD_SEND_REJECT_RESPONSE = "sendRejResp";

    /**
     * Команда отправки ответа со статусом
     */
    public static final String CMD_SEND_STATUS_RESPONSE = "sendStatusResp";

    /**
     * Команда получения статуса
     */
    public static final String CMD_GET_STATUS = "getStatus";

    /**
     * Команда получения статистики
     */
    public static final String CMD_GET_STATISTICS = "getStat";

    /**
     * Параметр командной строки: идентификатор узла
     */
    public static final String PARAM_NODE_ID = "nodeId";

    /**
     * Параметр командной строки: идентификатор сообщения, в ответ на которое отсылается текущее
     */
    public static final String PARAM_ORIGINAL_MESSAGE_ID = "origMsgId";

    /**
     * Параметр командной строки: MTOM вложения
     */
    public static final String PARAM_MTOM_ATTACHMENTS = "mtomAtts";

    /**
     * Параметр командной строки: большие вложения
     */
    public static final String PARAM_LARGE_ATTACHMENTS = "largeAtts";

    /**
     * Параметр командной строки: тип запроса
     */
    public static final String PARAM_REQUEST_TYPE_SELECTOR = "rts";

    /**
     * Параметр командной строки: код отказа обработки запроса
     */
    public static final String PARAM_CODE = "code";

    /**
     * Параметр командной строки: причина отказа обработки запроса
     */
    public static final String PARAM_DESCRIPTION = "desc";

    /**
     * Параметр командной строки: путь к файлу с бизнес-метаданными
     */
    public static final String PARAM_BPM = "bpm";

    /**
     * Параметр командной строки: дата завершения обслуживания
     */
    public static final String PARAM_EOL = "eol";

    /**
     * Параметр командной строки: файл с персональной электронной подписью
     */
    public static final String PARAM_PERSONAL_SIGNATURE = "ps";

    /**
     * Параметр командной строки: файл с дополнительными свойствами
     */
    public static final String PARAM_MAP = "map";

    /**
     * Параметр командной строки: признак тестового сообщения
     */
    public static final String PARAM_TEST_MESSAGE = "testMessage";

    /**
     * Шаблон взаимодействия со смэв
     */
    private Template template;

    /**
     * Запуск утилиты
     *
     * @param args параметры командной строки
     */
    public static void main(String[] args) {
        Main tool = null;
        try {
            tool = new Main(args);
            tool.execute();
        } catch (ParseException pe) {
            if (tool != null) {
                tool.printUsage();
            }
            LOG.error("Ошибка аргументов командной строки", pe);
            System.exit(1);
        } catch (Exception ex) {
            LOG.error("Ошибка утилиты", ex);
            System.exit(1);
        }
    }

    /**
     * Создает экземпляр утилиты с указанными параметрам
     *
     * @param args параметры командной строки
     */
    private Main(String[] args) throws SMEVRuntimeException {
        super(args);
        GeoTemplateConfiguration geoConfig = ToolConfiguration.getDefault().getGeoTemplateConfiguration();
        if (geoConfig != null) {
            template = Factory.getTemplateInstance(geoConfig);
        } else {
            template = Factory.getTemplateInstance(ToolConfiguration.getDefault().getTemplateConfiguration());
        }
    }

    /**
     * Исполняет команду, указанную в командной строке
     *
     * @throws Exception в случа возникновения ошибки
     */
    private void execute() throws Exception {
        String cmd = commandLine.getOptionValue(PARAM_CMD);
        String nodeId = commandLine.getOptionValue(PARAM_NODE_ID);
        String messageId = commandLine.getOptionValue(PARAM_MESSAGE_ID);
        String txId = commandLine.getOptionValue(PARAM_TRANSACTION_ID);
        String origMsgId = commandLine.getOptionValue(PARAM_ORIGINAL_MESSAGE_ID);
        String refMsgId = commandLine.getOptionValue(PARAM_REFERENCE_MESSAGE_ID);
        List<File> mtomAttachments = getFileListOptionValue(PARAM_MTOM_ATTACHMENTS);
        List<File> largeAttachments = getFileListOptionValue(PARAM_LARGE_ATTACHMENTS);
        QName requestTypeSelector = getQNameOptionValue(PARAM_REQUEST_TYPE_SELECTOR);
        String replyTo = commandLine.getOptionValue(PARAM_REPLY_TO);

        SMEVMetadata metadata = null;
        SMEVMessage message = null;
        if (cmd.equals(CMD_SEND_REQUEST)) {
            requireArgs(PARAM_TRANSACTION_ID, PARAM_INPUT_FILE);
            metadata = template.sendRequestMessage(nodeId, messageId, txId, origMsgId, refMsgId,
                    getBPM(), getDateArg(PARAM_EOL), fileToElement(getInputFile(commandLine)),
                    getPersonalSignature(), mtomAttachments, largeAttachments, getMapArg(PARAM_MAP), getBooleanOption (PARAM_TEST_MESSAGE));
        } else if (cmd.equals(CMD_GET_REQUEST)) {
            message = template.getRequestMessage(requestTypeSelector, nodeId);
        } else if (cmd.equals(CMD_SEND_RESPONSE)) {
            requireArgs(PARAM_REPLY_TO, PARAM_INPUT_FILE);
            metadata = template.sendResponseMessage(messageId, replyTo, origMsgId, getBPM(),
                    fileToElement(getInputFile(commandLine)), getPersonalSignature(), mtomAttachments, largeAttachments, getMapArg(PARAM_MAP));
        } else if (cmd.equals(CMD_GET_RESPONSE)) {
            message = template.getResponseMessage(requestTypeSelector, nodeId);
        } else if (cmd.equals(CMD_ACK)) {
            requireArgs(PARAM_MESSAGE_ID);
            template.ack(messageId);
            return;
        } else if (cmd.equals(CMD_SEND_REJECT_RESPONSE)) {
            requireArgs(PARAM_REPLY_TO, PARAM_CODE, PARAM_DESCRIPTION);
            metadata = template.sendRejectResponseMessage(messageId, replyTo, origMsgId,
                    commandLine.getOptionValue(PARAM_CODE), commandLine.getOptionValue(PARAM_DESCRIPTION), getBPM());
        } else if (cmd.equals(CMD_GET_STATUS)) {
            message = template.getStatusMessage(requestTypeSelector, nodeId);
        } else if (cmd.equals(CMD_SEND_STATUS_RESPONSE)) {
            requireArgs(PARAM_REPLY_TO, PARAM_CODE, PARAM_DESCRIPTION);
            metadata = template.sendStatusResponseMessage(messageId, replyTo, origMsgId,
                    getIntArg(PARAM_CODE), commandLine.getOptionValue(PARAM_DESCRIPTION), getMapArg(PARAM_MAP), getBPM());
        } else if (cmd.equals(CMD_GET_STATISTICS)) {
//            Map<String, Map<String, Long>> statistics = template.getIncomingQueueStatistics(nodeId);
//            writeOutput(statisticsToString2(statistics), null);
            return;
        } else {
            throw new ParseException("Неизвестная операция: " + cmd);
        }
        writeOutputData(message, metadata);
    }

    private boolean getBooleanOption (String optionName) {
        return "1".equals (commandLine.getOptionValue(optionName));
    }

    private List<Element> getBPM() throws Exception {
        List<File> files = getFileListOptionValue(PARAM_BPM);
        if ((files == null) || files.isEmpty()) {
            return null;
        }
        List<Element> result = new ArrayList<>(files.size());
        for (File file : files) {
            result.add(fileToElement(file));
        }
        return result;
    }

    private Element getPersonalSignature() throws Exception {
        if (!commandLine.hasOption(PARAM_PERSONAL_SIGNATURE)) {
            return null;
        }
        File f = new File(commandLine.getOptionValue(PARAM_PERSONAL_SIGNATURE));
        return fileToElement(f);
    }

    /**
     * Сохраняет результаты операции в файлы
     *
     * @param message  полученное сообщение (для get* операций)
     * @param metadata метаданные отправки (для send* операций)
     * @throws IOException в случае ошибки ввода-вывода
     */
    private void writeOutputData(SMEVMessage message, SMEVMetadata metadata) throws IOException {
        Properties meta = metaDataToProps(message == null ? metadata : message.getSMEVMetadata());
        addAttachments(meta, message);
        if ((message == null) || (message.getData() == null)) {
            writeOutput("NO_MESSAGE", null);
        } else {
            BusinessStatus status = message.getData().getStatus();
            if (status != null) {
                meta.setProperty("business.status", status.toString());
                switch (status) {
                    case REQUEST:
                    case RESPONSE:
                        writeOutput(message.getData().getContent(), null);
                        break;
                    case REJECT:
                        RejectResponseContent rrbc = (RejectResponseContent) message.getData();
                        for (RejectResponseContent.Reject reject : rrbc.getRejects()) {
                            if (reject.getStatusCode() != null) {
                                meta.setProperty(PARAM_CODE, reject.getStatusCode().toString());
                            }
                            if (reject.getStatusDescription() != null) {
                                meta.setProperty(PARAM_DESCRIPTION, reject.getStatusDescription());
                            }
                        }
                        break;
                    case STATUS:
                        StatusResponseContent srbc = (StatusResponseContent) message.getData();
                        meta.setProperty("statusCode", String.valueOf(srbc.getStatusCode()));
                        if (srbc.getStatusDescription() != null) {
                            meta.setProperty("statusDesc", srbc.getStatusDescription());
                        }
                        break;
                }
            }
            writeBPM(message.getData().getBusinessProcessMetadata());
        }
        writeOutput(meta, ".meta");
    }

    private void writeBPM(List<Element> bpm) throws IOException {
        if (bpm == null) {
            return;
        }
        int index = 0;
        for (Element element : bpm) {
            writeOutput(element, ".bpm" + (++index));
        }
    }

    /**
     * Конвертирует основные метаданные в мапу
     *
     * @param metadata метаданные сообщения
     * @return основные метаданные
     */
    private Properties metaDataToProps(SMEVMetadata metadata) {
        Properties result = new Properties();
        if (metadata == null) {
            return result;
        }

        SMEVMetadata.MessageIdentity mi = metadata.getMessageIdentity();
        if (mi != null) {
            if (mi.getMessageId() != null) {
                result.setProperty(PARAM_MESSAGE_ID, mi.getMessageId());
            }
            if (mi.getReferenceMessageId() != null) {
                result.setProperty(PARAM_REFERENCE_MESSAGE_ID, mi.getReferenceMessageId());
            }
            if (mi.getTransactionCode() != null) {
                result.setProperty(PARAM_TRANSACTION_ID, mi.getTransactionCode());
            }
        }

        SMEVContext context = metadata.getSmevContext();
        if (context != null) {
            if (context.getReplyTo() != null) {
                result.setProperty("replyTo", context.getReplyTo());
            }
            if (context.getSmevStatus() != null) {
                result.setProperty("smev.status", context.getSmevStatus().toString());
            }
            setExceptionDataToProps(context.getException(), result);
        }
        return result;
    }

    private void setExceptionDataToProps(SMEVException smevException, Properties props) {
        if (smevException instanceof SMEVProcessingException) {
            final SMEVProcessingException smevProcessingException = (SMEVProcessingException) smevException;
            props.setProperty("smev.exception.code", smevProcessingException.getCode());
            props.setProperty("smev.exception.description", smevProcessingException.getMessage());
            final Throwable exceptionCause = smevProcessingException.getCause();
            if (exceptionCause instanceof SMEVResourceException.DestinationOverflowException) {
                final SMEVResourceException.DestinationOverflowException.DestinationOverflow destinationOverflow
                        = ((SMEVResourceException.DestinationOverflowException) smevProcessingException.getCause()).getFaultInfo();
                props.setProperty("smev.exception.destination.name", destinationOverflow.getDestinationName());
                props.setProperty("smev.exception.message.broker.address", destinationOverflow.getMessageBrokerAddress());
            } else if (exceptionCause instanceof SMEVMessageContentException.BusinessDataTypeIsNotSupportedException) {
                final SMEVMessageContentException.BusinessDataTypeIsNotSupportedException.BusinessDataTypeIsNotSupported
                        faultInfo = ((SMEVMessageContentException.BusinessDataTypeIsNotSupportedException)
                        exceptionCause).getFaultInfo();
                props.setProperty("smev.exception.root.element.local.name", faultInfo.getRootElementLocalName());
                props.setProperty("smev.exception.root.element.namespace.URI", faultInfo.getRootElementNamespaceURI());
            } else if (exceptionCause instanceof SMEVMessageContentException.InvalidContentException) {
                final SMEVMessageContentException.InvalidContentException.InvalidContent faultInfo =
                        ((SMEVMessageContentException.InvalidContentException) exceptionCause).getFaultInfo();
                int valErrCount = 1;
                for (SMEVMessageContentException.InvalidContentException.InvalidContent.ValidationError valErr : faultInfo.getValidationError()) {
                    props.setProperty("smev.exception.validation.error." + valErrCount + ".value", valErr.getValue());
                    props.setProperty("smev.exception.validation.error." + valErrCount + ".position", valErr.getValue());
                    valErrCount++;
                }
            } else if (exceptionCause instanceof SMEVAttachmentsException.QuoteLimitExceededException) {
                final SMEVAttachmentsException.QuoteLimitExceededException.QuoteLimitExceeded faultInfo =
                        ((SMEVAttachmentsException.QuoteLimitExceededException) exceptionCause).getFaultInfo();
                props.setProperty("smev.exception.real.total.attachment.size", Long.valueOf(faultInfo
                        .getRealTotalAttachmentSize()).toString());
                props.setProperty("smev.exception.remaining.total.quote.size", Long.valueOf(faultInfo
                        .getRemainedTotalQuoteSize()).toString());
            } else if (exceptionCause instanceof SMEVSignatureException.SignatureVerificationFaultException) {
                props.setProperty("smev.exception.signature.fault",
                        ((SMEVSignatureException.SignatureVerificationFaultException) exceptionCause).getFaultInfo());
            }
        }
    }

    private void addAttachments(Properties meta, SMEVMessage message) {
        if ((message == null) || (message.getData() == null) || (message.getData().getAttachments() == null)) {
            return;
        }
        StringBuilder mtoms = new StringBuilder();
        StringBuilder large = new StringBuilder();
        for (SMEVAttachment attachment : message.getData().getAttachments()) {
            if (attachment instanceof MTOMAttachment) {
                if (mtoms.length() > 0) {
                    mtoms.append(";");
                }
                mtoms.append(((MTOMAttachment) attachment).getAttachmentId());
            } else if (attachment instanceof LargeAttachment) {
                if (large.length() > 0) {
                    large.append(";");
                }
                large.append(((LargeAttachment) attachment).getContentRef().getRawPath());
            }
        }
        if (mtoms.length() > 0) {
            meta.setProperty(PARAM_MTOM_ATTACHMENTS, mtoms.toString());
        }
        if (large.length() > 0) {
            meta.setProperty(PARAM_LARGE_ATTACHMENTS, large.toString());
        }
    }

    /**
     * Конвертирует указанную мапу в читаемую строку
     *
     * @param statistics статистика
     * @return статистика в виде строки
     */
    private String statisticsToString(Map<String, Long> statistics) {
        StringBuilder result = new StringBuilder();
        for (Map.Entry<String, Long> entry : statistics.entrySet()) {
            if (result.length() > 0) {
                result.append("\n");
            }
            result.append(entry.getKey()).append(" = ").append(entry.getValue());
        }
        return result.toString();
    }

    private String statisticsToString2(Map<String, Map<String, Long>> statistics) {
        StringBuilder builder = new StringBuilder();
        for (Map.Entry<String, Map<String, Long>> entry : statistics.entrySet()) {
            if (builder.length() > 0) {
                builder.append("\n");
            }
            builder.append(entry.getKey()).append(":\n");
            builder.append(statisticsToString(entry.getValue()));
        }
        return builder.toString();
    }

    /**
     * Добавляет специфические для данной утилиты параметры
     *
     * @param options описание параметров командной строки
     */
    @Override
    protected void populateOptions(Options options) {
        //команды
        String cmdDesc = String.format("Выполняемая операция. Возможные значения: %s - отправка запроса, " +
                        "%s - получение запроса, %s - отправка ответа, %s - получение ответа, " +
                        "%s - подтверждение получения сообщения, %s - отклонение запроса, " +
                        "%s - получение статуса, %s - отправка ответа со статусом, " +
                        "%s - получение статистики",
                CMD_SEND_REQUEST, CMD_GET_REQUEST, CMD_SEND_RESPONSE, CMD_GET_RESPONSE, CMD_ACK, CMD_SEND_REJECT_RESPONSE,
                CMD_GET_STATUS, CMD_SEND_STATUS_RESPONSE, CMD_GET_STATISTICS);
        Option cmd = Option.builder(PARAM_CMD).hasArg().required().argName("тип операции").desc(cmdDesc).build();
        options.addOption(cmd);
        //параметры
        Option inputFile = Option.builder(PARAM_INPUT_FILE).hasArg().argName("имя файла").desc("Путь к входному файлу с содержанием сообщения").build();
        options.addOption(inputFile);
        Option outputFile = Option.builder(PARAM_OUTPUT_FILE).hasArg().argName("имя файла").desc("Путь к выходному файлу").build();
        options.addOption(outputFile);
        Option transactionId = Option.builder(PARAM_TRANSACTION_ID).hasArg().argName("идентификатор").desc("Идентификатор транзакции").build();
        options.addOption(transactionId);
        Option nodeId = Option.builder(PARAM_NODE_ID).hasArg().argName("идентификатор").desc("идентификатор узла").build();
        options.addOption(nodeId);
        Option msgId = Option.builder(PARAM_MESSAGE_ID).hasArg().argName("идентификатор").desc("идентификатор сообщения").build();
        options.addOption(msgId);
        Option origMsgId = Option.builder(PARAM_ORIGINAL_MESSAGE_ID).hasArg().argName("идентификатор").desc("Идентификатор сообщения, в ответ на которое отсылается текущее").build();
        options.addOption(origMsgId);
        Option refMsgId = Option.builder(PARAM_REFERENCE_MESSAGE_ID).hasArg().argName("идентификатор").desc("Идентификатор изначального сообщения").build();
        options.addOption(refMsgId);
        Option mtomAtts = Option.builder(PARAM_MTOM_ATTACHMENTS).hasArg().argName("список файлов").desc("Файлы вложений, разделенные точкой с запятой").build();
        options.addOption(mtomAtts);
        Option largeAtts = Option.builder(PARAM_LARGE_ATTACHMENTS).hasArg().argName("список файлов").desc("Файлы вложений большого размера, разделенные точкой с запятой").build();
        options.addOption(largeAtts);
        Option rts = Option.builder(PARAM_REQUEST_TYPE_SELECTOR).hasArg().argName("имя").desc("Квалифицированное имя XML получаемого запроса").build();
        options.addOption(rts);
        Option replyTo = Option.builder(PARAM_REPLY_TO).hasArg().argName("идентификатор").desc("Адрес получателя ответа").build();
        options.addOption(replyTo);
        Option code = Option.builder(PARAM_CODE).hasArg().argName("код").desc("Код-причина отклонения запроса").build();
        options.addOption(code);
        Option description = Option.builder(PARAM_DESCRIPTION).hasArg().argName("описание").desc("Описание причины отклонения запроса").build();
        options.addOption(description);
        Option bpm = Option.builder(PARAM_BPM).hasArg().argName("список файлов").desc("Файлы с бизнес-метаданными, разделенные точкой с запятой").build();
        options.addOption(bpm);
        Option eol = Option.builder(PARAM_EOL).hasArg().argName("дата").desc("Дата завершения обслуживания в формате ДД.ММ.ГГГГ ЧЧ:ММ:СС").build();
        options.addOption(eol);
        Option ps = Option.builder(PARAM_PERSONAL_SIGNATURE).hasArg().argName("имя файла").desc("Путь к файлу с персональной подписью").build();
        options.addOption(ps);
        Option map = Option.builder(PARAM_MAP).hasArg().argName("имя файла").desc("Путь к файлу с дополнительными свойствами").build();
        options.addOption(map);
        Option testMessage = Option.builder(PARAM_TEST_MESSAGE).hasArg().argName("1 или 0").desc("Признак тестового сообщения").build();
        options.addOption(testMessage);
    }

    /**
     * Возвращает нижнюю подпись к помощи по командной строке
     *
     * @return подпись к помощи по командной строке
     */
    @Override
    protected String getHelpFooter() {
        StringBuilder str = new StringBuilder();
        str.append("Возможные параметры для команд:\n");
        str.append("-cmd sendReq: -tid -if [-nodeId -msgId -origMsgId -refMsgId -bpm -eol -ps -mtomAtts -largeAtts -map -testMessage]\n");
        str.append("-cmd getReq: [-nodeId -rts]\n");
        str.append("-cmd sendResp: -replyTo -if [-msgId -origMsgId -bpm -ps -mtomAtts -largeAtts -map]\n");
        str.append("-cmd getResp: [-nodeId -rts]\n");
        str.append("-cmd ack: -msgId\n");
        str.append("-cmd sendRejResp: -replyTo -code -desc [-msgId -origMsgId -bpm]\n");
        str.append("-cmd getStatus: [-nodeId -rts]\n");
        str.append("-cmd sendStatusResp: -replyTo -code -desc [-msgId -origMsgId -bpm -map]\n");
        str.append("-cmd getStat: [-nodeId]\n");
        str.append("[...] - необязательные параметры");
        return str.toString();
    }
}
