package ru.voskhod.smev.client.toolset.signer;

import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.signature.Signer;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.api.types.exception.processing.SMEVSignatureException;
import ru.voskhod.smev.client.toolset.commons.AbstractTool;
import ru.voskhod.smev.client.toolset.commons.ToolConfiguration;

import javax.xml.bind.DatatypeConverter;
import java.io.File;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;

/**
 * Утилита для использования {@link ru.voskhod.smev.client.api.services.signature.Signer крипто-сервиса}
 */
public class Main extends AbstractTool {

    /**
     * Операция получения цифровой подписи XML файла
     */
    public static final String CMD_SIGN_XML = "signXml";

    /**
     * Операция получения цифровой подписи двоичного хэша файла
     */
    public static final String CMD_SIGN_HASH = "signHash";

    /**
     * Операция получения хэша
     */
    public static final String CMD_DIGEST = "digest";

    /**
     * Операция проверки подписи СМЭВ
     */
    public static final String CMD_VALIDATE = "validate";

    /**
     * Параметр командной строки: файл с подписью СМЭВ
     */
    public static final String PARAM_SMEV_SIGNATURE = "sig";

    /**
     * Логгер
     */
    private static final Logger LOG = LoggerFactory.getLogger(Main.class);

    /**
     * Крипто-сервис
     */
    private Signer signer;

    /**
     * Запускает утилиту
     *
     * @param args параметры командной строки
     */
    public static void main(String[] args) {
        try {
            new Main(args).execute();
        } catch (Exception ex) {
            LOG.error("Ошибка утилиты", ex);
            System.exit(1);
        }
    }

    /**
     * Создает экземпляр утилиты с указанными параметрам
     *
     * @param args параметры командной строки
     */
    private Main(String[] args) throws SMEVRuntimeException {
        super(args);
        signer = Factory.getSignerInstance(ToolConfiguration.getDefault().getSignerConfiguration());
    }

    /**
     * Исполняет команду, указанную в командной строке
     *
     * @throws Exception в случа возникновения ошибки
     */
    private void execute() throws Exception {
        String cmd = commandLine.getOptionValue(PARAM_CMD);
        if (cmd.equals(CMD_SIGN_XML)) {
            Element result = signer.sign(fileToElement(getInputFile(commandLine)));
            writeOutput(result, null);
        } else if (cmd.equals(CMD_SIGN_HASH)) {
            byte[] hash = DatatypeConverter.parseBase64Binary
                    (new String(Files.readAllBytes(getInputFile(commandLine).toPath()), StandardCharsets.UTF_8.name()));
            writeOutput(DatatypeConverter.printBase64Binary(signer.signPKCS7Detached(hash)), null);
        } else if (cmd.equals(CMD_DIGEST)) {
            writeOutput(DatatypeConverter.printBase64Binary(signer.getDigest(getInputFile(commandLine))), null);
        } else if (cmd.equals(CMD_VALIDATE)) {
            validateSmevSignature();
        } else {
            throw new ParseException("Неизвестная операция: " + cmd);
        }
    }

    /**
     * Добавляет специфические для данной утилиты параметры
     *
     * @param options описание параметров командной строки
     */
    @Override
    protected void populateOptions(Options options) {
        //команды
        String cmdDesc = String.format("Выполняемая операция. Возможные значения: %s - получить цифровую подпись xml-файла, " +
                        "%s - получить цифровую подпись двоичного хэша файла, " +
                        "%s - получить хэш файла, " +
                        "%s - проверить цифровую подпись СМЭВ",
                CMD_SIGN_XML, CMD_SIGN_HASH, CMD_DIGEST, CMD_VALIDATE);
        Option cmd = Option.builder(PARAM_CMD).hasArg().required().argName("тип операции").desc(cmdDesc).build();
        options.addOption(cmd);
        //параметры
        Option inputFile = Option.builder(PARAM_INPUT_FILE).required().hasArg().argName("имя файла").desc("Путь к входному файлу с бинарным хэшем или XML данными").build();
        options.addOption(inputFile);
        Option outputFile = Option.builder(PARAM_OUTPUT_FILE).hasArg().argName("имя файла").desc("Путь к файлу результата. В случае отсутствия данного параметра, вывод осуществляется в консоль.").build();
        options.addOption(outputFile);
        Option smevSignature = Option.builder(PARAM_SMEV_SIGNATURE).hasArg().argName("имя файла").desc("Путь к файлу с подписью СМЭВ. Только для команды " + CMD_VALIDATE).build();
        options.addOption(smevSignature);
    }

    private void validateSmevSignature() throws Exception {
        if (!commandLine.hasOption(PARAM_SMEV_SIGNATURE)) {
            throw new ParseException("Файл с подписью не указан");
        }
        Element signature = fileToElement(new File(commandLine.getOptionValue(PARAM_SMEV_SIGNATURE)));
        Element content = fileToElement(getInputFile(commandLine));
        try {
            signer.validateSMEVSignature(signature, content);
            writeOutput("true", null);
        } catch (SMEVSignatureException ex) {
            writeOutput("false", null);
        }
    }

}
