package ru.voskhod.smev.client.toolset.transaction;

import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.voskhod.smev.client.api.factory.Factory;
import ru.voskhod.smev.client.api.services.transaction.TransactionService;
import ru.voskhod.smev.client.api.types.exception.SMEVRuntimeException;
import ru.voskhod.smev.client.toolset.commons.AbstractTool;
import ru.voskhod.smev.client.toolset.commons.ToolConfiguration;

/**
 * Утилита для использования {@link ru.voskhod.smev.client.api.services.transaction.TransactionService сервиса транзакций}
 */
public class Main extends AbstractTool {

    /**
     * Параметр командной строки: Код ФРГУ информационной системы
     */
    public static final String PARAM_SERVICE_CODE = "serviceCode";

    /**
     * Параметр командной строки: Код ФРГУ услуги либо признака функции
     */
    public static final String PARAM_PARTICIPANT_CODE = "participantCode";

    /**
     * Параметр командной строки: Расширенные сведения об услуге либо функции
     */
    public static final String PARAM_SERVICE_DESCR = "serviceDescr";

    /**
     * Параметр командной строки: Расширенные сведения о получателе услуги либо функции
     */
    public static final String PARAM_PARTICIPANT_DESCR = "participantDescr";

    /**
     * Логгер
     */
    private static final Logger LOG = LoggerFactory.getLogger(Main.class);

    /**
     * Сервис генерации идентификаторов транзакций
     */
    private TransactionService service;

    /**
     * Генерирует идентификатор транзакции по указанным параметрам
     *
     * @param args параметры командной строки
     */
    public static void main(String[] args) {
        try {
            new Main(args).execute();
        } catch (Exception ex) {
            LOG.error("Ошибка утилиты", ex);
            System.exit(1);
        }
    }

    /**
     * Создает экземпляр утилиты с указанными параметрам
     *
     * @param args параметры командной строки
     */
    private Main(String[] args) throws SMEVRuntimeException {
        super(args);
        service = Factory.getTransactionService(ToolConfiguration.getDefault().getTxConfiguration());
    }

    /**
     * Исполняет команду, указанную в командной строке
     *
     * @throws Exception в случа возникновения ошибки
     */
    private void execute() throws Exception {
        String code = service.getTransactionCode(
                commandLine.getOptionValue(PARAM_SERVICE_CODE),
                commandLine.getOptionValue(PARAM_PARTICIPANT_CODE),
                commandLine.getOptionValue(PARAM_SERVICE_DESCR),
                commandLine.getOptionValue(PARAM_PARTICIPANT_DESCR));
        writeOutput(code, null);
    }

    /**
     * Добавляет специфические для данной утилиты параметры
     *
     * @param options описание параметров командной строки
     */
    @Override
    protected void populateOptions(Options options) {
        Option serviceCode = Option.builder(PARAM_SERVICE_CODE).hasArg().argName("КОД").required().desc("Код ФРГУ информационной системы").build();
        options.addOption(serviceCode);
        Option participantCode = Option.builder(PARAM_PARTICIPANT_CODE).hasArg().argName("КОД").required().desc("Код ФРГУ услуги либо признака функции").build();
        options.addOption(participantCode);
        Option serviceDescr = Option.builder(PARAM_SERVICE_DESCR).hasArg().argName("ОПИСАНИЕ").desc("Расширенные сведения об услуге либо функции").build();
        options.addOption(serviceDescr);
        Option participantDescr = Option.builder(PARAM_PARTICIPANT_DESCR).hasArg().argName("ОПИСАНИЕ").desc("Расширенные сведения о получателе услуги либо функции").build();
        options.addOption(participantDescr);
        Option outputFile = Option.builder(PARAM_OUTPUT_FILE).hasArg().argName("имя файла").desc("Путь к выходному файлу").build();
        options.addOption(outputFile);
    }
}
